import numpy as np
from . import util
import scipy.spatial as spatial

class BinnedKDTrees(object):
    def __init__(self, x, mpeak, ok, bin_factor=1.25):
        log_mpeak = np.log10(mpeak)
        log_m_min, log_m_max = np.min(log_mpeak), np.max(log_mpeak)
        d_log_m = np.log10(bin_factor)

        n_bins = int(np.ceil((log_m_max - log_m_min) / d_log_m))

        self.bins = np.linspace(log_m_min, log_m_min+(n_bins+1)*d_log_m, n_bins+1)
        self.trees = [None]*n_bins
        self.idxs = [None]*n_bins

        for i in range(n_bins):
            ok_i = ok & (log_mpeak >= self.bins[i]) & (log_mpeak < self.bins[i+1])
            if np.sum(ok_i) > 0:
                self.idxs[i] = np.where(ok_i)[0]
                self.trees[i] = spatial.KDTree(x[ok_i])

    def best_match(self, x, mpeak, search_factor=2):
        m_low = np.log10(mpeak/search_factor)
        m_high = np.log10(mpeak*search_factor)
        
        idx_high = min(np.searchsorted(self.bins, m_high), len(self.bins)-1)
        idx_low = max(np.searchsorted(self.bins, m_low), 0)

        match_idxs, match_rs = [], []

        for i in range(idx_low, idx_high):
            if self.trees[i] is None: continue
            r, j = self.trees[i].query(x)
            match_rs.append(r)
            match_idxs.append(self.idxs[i][j])

        if len(match_rs) == 0:
            return -1

        return match_idxs[np.argmin(match_rs)]
        

def match_subhalos(h_1, h_2, hist_1, hist_2, min_votes=4):
    """ match_subhalos matches subhalos from simulation 1 against simulation 2
    using their pre-infall trajectories. h_1 and h_2 are 2D ROCKSTAR_DTYPE
    arrays, as returned by read_rockstar(). Any similarly structured
    2D array will work as long as it has a position field called "x" and a
    field called "ok" that indicates whether the subhalo exists. hist_1
    and hist_2 are HISTORY_DYTPE arrays, as returned by read_rockstar(). Any
    similarly formatted strucutred array will work as long as it also has
    a pre-infall mpeak field called "mpeak_pre" and an infall snapshot field
    called "first_infall_snap".

    This can be done with either comoving positions or physical displacements
    from the central halo, but we've noticed that the matching is usually
    more stable to chaotic changes in host position if comoving units are used.

    Two arrays are returned, match_1 and match_2, each of length |hist_1| and
    |hist_2|, respectively. They give the index of the corresponding match in
    the other simulation. this index will be unique and can be -1 if no
    reliable match can be found.
    
    The optional argument min_votes can be used to tune the matching. Higher
    values lead to fewer false positives and lower values lead to fewer false
    positives. 4 works well for SyphonyMilkyWay, but if looking at a simulation
    with a different output cadence, you should experiment with this. The only
    good way to test this that we are aware of is to qualitatively look at
    MAHs and movies, such as through the use of plot_matched_subhalos().
    """
    n_snap = h_1.shape[1]
    mpeak_1 = hist_1["mpeak_pre"]
    mpeak_2 = hist_2["mpeak_pre"]

    votes_1, votes_2 = [], []

    for snap in range(n_snap):
        match_1 = np.ones(len(h_1), dtype=int)*-1
        match_2 = np.ones(len(h_2), dtype=int)*-1

        last_snap_1 = hist_1["first_infall_snap"]
        last_snap_2 = hist_2["first_infall_snap"]
        last_snap_1[last_snap_1 < 0] = n_snap+1
        last_snap_2[last_snap_2 < 0] = n_snap+1

        ok_1 = (snap < hist_1["first_infall_snap"]) & h_1["ok"][:,snap]
        ok_2 = (snap < hist_2["first_infall_snap"]) & h_2["ok"][:,snap]
        trees_1 = BinnedKDTrees(h_1["x"][:,snap,:], mpeak_1, ok_1)
        trees_2 = BinnedKDTrees(h_2["x"][:,snap,:], mpeak_2, ok_2)

        for i in np.where(ok_1)[0]:
            match_1[i] = trees_2.best_match(h_1["x"][i,snap], mpeak_1[i])
        for i in np.where(ok_2)[0]:
            match_2[i] = trees_1.best_match(h_2["x"][i,snap], mpeak_2[i])
                    
        for i in range(len(match_1)):
            if match_2[match_1[i]] == i:
                votes_1.append(i)
                votes_2.append(match_1[i])

    votes_1 = np.array(votes_1, dtype=int)
    votes_2 = np.array(votes_2, dtype=int)        

    vote_count = { }

    for i in range(len(votes_1)):
        pair = (votes_1[i], votes_2[i])
        vote_count[pair] = 1 + vote_count.get(pair, 0)

    idx_1, idx_2, n_votes = process_vote_count(vote_count)
    order = np.argsort(n_votes)[::-1]
    idx_1, idx_2, n_votes = idx_1[order], idx_2[order], n_votes[order]

    match_1 = np.ones(len(h_1), dtype=int)*-1
    match_2 = np.ones(len(h_2), dtype=int)*-1
    in_use_1, in_use_2 = {}, {}
    for i in range(len(n_votes)):
        if idx_1[i] in in_use_1 or idx_2[i] in in_use_2 or n_votes[i] < min_votes:
            continue
        i1, i2 = idx_1[i], idx_2[i]
        match_1[i1] = i2
        match_2[i2] = i1
        in_use_1[i1] = None
        in_use_2[i2] = None
    
    return match_1, match_2
    
def process_vote_count(vote_count):
    pairs = vote_count.keys()
    idx_1, idx_2 = list(zip(*pairs))
    idx_1, idx_2 = np.array(idx_1, dtype=int), np.array(idx_2, dtype=int)
    n_votes = np.array([vote_count[pair] for pair in pairs], dtype=int)
    return idx_1, idx_2, n_votes

def plot_matched_subhalos(ax, h_1, h_2, match_1, match_2, snap, r_max,
                          plot_set_1=None, plot_set_2=None):
    """ plot_matched_subhalos plots plots the connections between subhalos
    according to the results of match_subhalos(). ax is the pyplot axis that
    will be used, h_1 and h_2 are 2D ROCKSTAR_DTYPE arrays from the matched
    simulations, formatted as returned by read_rockstar(). match_1 and match_2
    are the results of match_subhalos(), snap is the snapshot being plotted,
    and r_max is the maximum radius to plot relative to the host center, given
    in physical kpc.

    Two optional argument, plot_set_1 and plot_set_2 can be provided. They give
    the indices of all the subhaloes to be plotted.

    Subhalos in sim 1 will be plotted with solid lines and subhalos in sim 2
    will be plotted as dashed lines. Mached subhalos will be plotted in blue
    with a line connecting them. Subhaloes which are matched to a subhalo that
    doesn't exist in the current snapshot will be red, and subhalos which are
    not matched to anything will be plotted in black.

    This is indended to be used to generate a single frame in a movie.
    """
    if plot_set_1 is None: plot_set_1 = np.arange(len(h_1), dtype=int)
    if plot_set_2 is None: plot_set_2 = np.arange(len(h_2), dtype=int)

    ls_1, ls_2 = "-", "--"
    main_c = "tab:gray"
    surv_c, dis_c, unmatch_c = "tab:blue", "tab:red", "black"

    for i in plot_set_1:
        if not h_1["ok"][i,snap]: continue

        m = match_1[i]

        if i == 0:
            c = main_c
        elif m == -1:
            c = unmatch_c
        elif not h_2[m,snap]["ok"]:
            c = dis_c
        else:
            c = surv_c
            
        util.plot_circle(ax, h_1["x"][i,snap,0],
                         h_1["x"][i,snap,1],
                         h_1["rvir"][i,snap],
                         c=c, ls=ls_1, lw=1.5)
        ax.plot([h_1["x"][i,snap,0]], [h_1["x"][i,snap,1]], ".", c=main_c)

        if c == surv_c:
            util.plot_circle(ax, h_2["x"][m,snap,0],
                             h_2["x"][m,snap,1],
                             h_2["rvir"][m,snap],
                             c=c, ls=ls_2, lw=1.5)
            ax.plot([h_2["x"][m,snap,0]], [h_2["x"][m,snap,1]], ".", c=main_c)
            ax.plot([h_2["x"][m,snap,0], h_1["x"][i,snap,0]],
                     [h_2["x"][m,snap,1], h_1["x"][i,snap,1]],
                     lw=1, c=main_c)

    for i in plot_set_2:
        if not h_2["ok"][i,snap]: continue

        m = match_2[i]

        if i == 0:
            c = main_c
        elif m == -1:
            c = unmatch_c
        elif not h_1[m,snap]["ok"]:
            c = dis_c            
        else:
            c = surv_c

        util.plot_circle(ax, h_2["x"][i,snap,0],
                           h_2["x"][i,snap,1],
                           h_2["rvir"][i,snap],
                           c=c, ls=ls_2, lw=1.5)
        ax.plot([h_2["x"][i,snap,0]], [h_2["x"][i,snap,1]], ".", c=main_c)

        if c == surv_c:
            util.plot_circle(ax, h_1["x"][m,snap,0],
                               h_1["x"][m,snap,1],
                               h_1["rvir"][m,snap],
                               c=c, ls=ls_1, lw=1.5)
            ax.plot([h_1["x"][m,snap,0]], [h_1["x"][m,snap,1]], ".", c=main_c)
            ax.plot([h_1["x"][m,snap,0], h_2["x"][i,snap,0]],
                    [h_1["x"][m,snap,1], h_2["x"][i,snap,1]],
                    lw=1, c=main_c)

    ax.set_xlim(-r_max, +r_max)
    ax.set_ylim(-r_max, +r_max)
    ax.set_xlabel(r"$X\ ({\rm kpc})$")
    ax.set_ylabel(r"$Y\ ({\rm kpc})$")



import array
import struct
import numpy as np
import os
import os.path as path
import scipy.interpolate as interpolate
import numpy.random as random
import scipy.optimize as optimize
import glob
from . import util

""" SUBHALO_DTYPE is the numpy datatype used by the main return value of
read_subhalos().
"""
SUBHALO_DTYPE = [("id", "i4"), ("mvir", "f4"), ("vmax", "f4"), ("rvmax", "f4"),
                 ("x", "f4", (3,)), ("v", "f4", (3,)), ("ok", "?"),
                 ("rvir", "f4"), ("cvir", "f4")]


""" ROCKSTAR_DTYPE is the main numpy datatype returned by read_rockstar().
 - id: unique identifier for the halo.
 - m (Msun): mass of the subhalo.
 - vmax (km/s): maximum value of the subhalo's rotation curve.
 - rvmax (physical kpc): radius of vmax.
 - x (physical kpc): displacement from the host's centre to the subhalo's
   centre.
 - v (km/s) velocity of the subhalo relative to the host's centre.
 - rvir (physical kpc): radius of the subhalo according to Bryan & Norman 1998.
   This is rarely a meaningful quantity for subhaloes as their tidal radii are
   usually smaller than rvir.
 - cvir: the NFW concentration parameter, r_-2/rvir for the subhalo, as
   estimated from vmax/vvir. This is rarely a meaninfiul quantity for
   subhaloes, because their profiles are very different from NFW..
 - ok: True if the subhalo exists and False otherwise. Note that if Rockstar
   has any errors but still believes this subhalo is real, this will be true.
   This can only be calculated for haloes which symfind has been run on
   You must use the HISTORY_DTYPE or SYMFIND_DTYPE to identify
"""
ROCKSTAR_DTYPE = [("id", "i4"), ("m", "f4"), ("vmax", "f4"), ("rvmax", "f4"),
                 ("x", "f4", (3,)), ("v", "f4", (3,)), ("ok", "?"),
                 ("rvir", "f4"), ("cvir", "f4")]

""" HISTORY_DTYPE is a numpy datatype representing data about a subhalo which 
is indepdent of time, such as the maximum mass that it takes on, its merger
snapshot, and its location in the full merger tree file.

 - The main branch of a given halo within the depth-first merger tree can be 
   found with x[start:end].
 - mpeak: (Msun) the peak value of Mvir across the entire branch
 - mpeak_pre: (Msun) the peak value of Mvir before becoming a subhalo
 - vpeak: (km/s) the peak value of vmax across the entire branch
 - merger_snap: the snapshot that the subhalo first became a subhalo of the
   main host halo.
 - first_infall_snap: the snapshot that the subhalo first became a subhalo of
   any other object, excluding glitches during major mergers.
 - conv_snap_*: the snaphot at which the subhalo is predicted to be
   unconverged, according to various criteria. See mansfield et al. 2023 for
   discussion
 - disrupt_snap: the snapshot at which the subhalo disrupts in the symfind
   catalogue
 - disrupt_snap_rs: the snapshot at which the subhalo disrupts in Rockstar + 
   consistent-trees. If the Rockstar has erroneous subhalo
   identifications at the end of its branch, this flags the last non-error
   snapshot.
 - several other non-informative fields are included for consistency with
   BRANCH_DTYPE.
"""
HISTORY_DTYPE = [("mpeak", "f4"), ("vpeak", "f4"), ("merger_snap", "i4"),
                 ("merger_ratio", "f4"),
                 ("start", "i4"), ("end", "i4"), ("is_real", "?"),
                 ("is_disappear", "?"), ("is_main_sub", "?"),
                 ("preprocess", "i4"), ("first_infall_snap", "i4"),
                 ("branch_idx", "i4"), ("false_selection", "?"),
                 ("mpeak_pre", "f4"), ("conv_snap_discrete", "i4"),
                 ("conv_snap_eps", "i4"), ("conv_snap_relax", "i4"),
                 ("conv_snap_relax_hydro", "i4"), ("disrupt_snap", "i4"),
                 ("disrupt_snap_rs", "i4")]


""" BRANCH_DTYPE is a numpy datatype representing the main branch of of halo's
merger tree. This forms a component of HISTORY_DTYPE, which 
 - The main branch of a given halo within the depth-first merger tree can be 
   found with x[start:end].
 - is_real: false if a halo is definitely a numerical artefact (e.g. if it's
   already a subhalo in its first snapshot).
 - is_disappear: true if the subhalo disappears without merging. This is also
   bad and probably means the thing is a numerical artefact.
 - is_main_sub: true is the halo was ever a subhalo of the zoom-in box's main 
   halo (includes splashback subhaloes).
 - preprocess: the index of the branch of the largest halo that hosted this 
   halo before it became a subhalo of the main halo. If the halo was never a
   subhalo of the main halo, this is just the index of largest halo to have
   every hosted this halo. If no other halo every hosted this halo before
   it entered Rvir of the main halo, this is -1. Includes splashback subhaloes
   and doesn't include collisions between subhaloes once they've already been
   accreted.
 - branch_idx: the index of this halo in the full merger tree.
"""
BRANCH_DTYPE = [("start", "i4"), ("end", "i4"), ("is_real", "?"),
                 ("is_disappear", "?"), ("is_main_sub", "?"),
                 ("preprocess", "i4"), ("first_infall_snap", "i4")]


""" CORE_DTYPE is a numpy datatype representing the properties of the
particle-tracked subhalo.
TODO: UPDATE THIS COMMENT
 - x: position in pkpc
 - v: velocity in km/s
 - r_tidal: tidal radius in pkpc. Accounts for angular momentum and the
   mass profile of the central halo.
 - r50_bound: radius enclosing 50% of bound particles
 - r50_bound_rs: radius enclosing 50% of bound particles
 - r95_bound: radius enclosing 95% of bound particles
 - m_tidal: total mass within tidal radius
 - m_tidal_bound: total bound mass within tidal radius
 - m_bound: total bound mass
 - ok: true if the core is being tracked and false otherwise
 - ok_rs: true if a rockstar halo exists and at least one core particle is
   still inside Rockstar's r_50 (almost all cases where this fails are cases
   where Rockstar has made an error)
 - is_err: Rockstar and core-tracking disagree, but more core particles are
   assoicated with Rockstar
 - is_err_rs: Rockstar and core-tracking disagree, but more core particles are
   with the particle-tracking subhalo.
 - interp: true if the core in this snapshot is interpolated betwen two
   nearby snapshots.
"""
CORE_DTYPE = [("x", "f4", (3,)), ("v", "f4", (3,)), ("r_tidal", "f4"),
              ("r50_bound", "f4"), ("r50_bound_rs", "f4"), ("m_tidal", "f4"),
              ("m_tidal_bound", "f4"), ("m_bound", "f4"), ("vmax", "f4"),
              ("f_core", "f4"), ("f_core_rs", "f4"), ("d_core_mbp", "f4"),
              ("ok", "?"), ("ok_rs", "?"), ("is_err", "?"), ("is_err_rs", "?"),
              ("interp", "?")]

""" SYMFIND_DTYPE is the main numpy datatype returned by read_symfind().
 - x (physical kpc): displacement from the host halo to the subhalo.
 - v (km/s): velocity relative to the host.
 - m (Msun): bound mass of the subhalo. Post-porcessing has been done to reduce
   noise by enfocing monotonicity (i.e. m(z) = max(m_raw(>= z)))
 - m_raw (Msun): bound mass of the suhalo with no post-processing.
 - r_half (physical kpc) half-mass radius of the subhalo.
 - r_tidal (physical kpc): tidal radius of the subhalo. Accounts for angular
   momentum and the mass profile of the central halo. Tidal radii can be
   poor approximations if the size of the subhalo is comparable to the
   distance to the host centre or if the subhalo is at pericentre. There is no
   strict criteria for identifying this.
 - m_tidal (Msun): total mass within tidal radius.
 - m_tidal_bound (Msun): total bound mass within tidal radius, using only
   particles within the tidal radius. This value can be a bit fragile at
   pericentre.
 - ok: true if the subhalo exists and is not an error, and false otherwise.
   Interpolated subhaloes will be marked as true.
 - interp: true if the subhalo in this snapshot is interpolated betwen two
   nearby snapshots.

Some values for Rockstar haloes are also calculated here:
 - r_half_rs (physical kpc): half-mass radius of tracked particles using
   Rockstar's position and velocity. This is the r_half used for error
   detection, not the r_half in the Rockstar catalogues.
 - ok_rs: true if a rockstar halo exists and at least one core particle is
   still inside Rockstar's r_half. A slightly more conservative way to
   identify Rockstar errors would be to check if ok_rs & (~is_err_rs).

Some diagnostic values are included that will not be useful for a typical user:
 - is_err: Rockstar and symfind disagree, but more core particles are
   assoicated with Rockstar. If this is true, the subhalo is either
   interpolated or flagged as disrupted.
 - is_err_rs: Rockstar and symfind disagree, but more core particles are
   associated with symfind.
 - f_core: the fraction of core particles associated with Rockstar
 - f_core_rs: the fraction of core particles associated with symfind.
 - d_core_mbp (physical kpc): the distance between the subhalo and the
   most-bound particle. Sometimes the most-bound particle can numerically
   diffuse out of the subhalo.
"""
SYMFIND_DTYPE = [("x", "f4", (3,)), ("v", "f4", (3,)), ("r_tidal", "f4"),
                 ("r_half", "f4"), ("m_tidal", "f4"),
                 ("m_tidal_bound", "f4"), ("m", "f4"), ("vmax", "f4"),
                 ("m_raw", "f4"),
                 ("ok", "?"), ("interp", "?"),
                 ("f_core", "f4"), ("f_core_rs", "f4"), ("d_core_mbp", "f4"),
                 ("r_half_rs", "f4"), ("ok_rs", "?"),
                 ("is_err", "?"), ("is_err_rs", "?")]

""" UM_DTYPE is a numpy datatype representing UniverseMachine predictions for
galaxy properties.
- m_star: the stellar mass of the galaxy in Msun
- m_in_situ: the stellar mass of a galaxy that formed inside the galaxy (as
             opposed to coming in via mergers) in Msun
- m_icl: the stellar mass of the galaxy's stellar halo
- sfr: the starformaiton rate in Msun/yr
- mvir: predicted virial mass of the halo in msun
- vmax: predicted V_max of the halo in km/s
- x: position of the galaxy relative to the host halo in pkpc
- v: velocity of the galaxy relative to the host halo in km/s
- rank: the galaxy's rank in delta vmax
- is_orphan: true if the the galaxy is un-tracked and replaced with an orphan
             model during this timestep and false otherwise
- ok: true if UniverseMachine has an entry for this galaxy at this snapshot and
      false otherwise
"""
UM_DTYPE = [("m_star", "f4"), ("m_in_situ", "f4"), ("m_icl", "f4"),
            ("sfr", "f4"), ("x", "f4", (3,)), ("v", "f4", (3,)),
            ("rank", "f4"), ("mvir", "f4"), ("vmax", "f4"),
            ("is_orphan", "?"), ("ok", "?")]

GALAXY_DTYPE = [("m_star", "f4"), ("r_half", "f4"), ("m_dyn", "f4"),
                ("v_disp_3d_dm", "f4"), ("v_disp_3d_star", "f4"),
                ("vmax_dm", "f4"), ("vmax_dm_debias", "f4",),
                ("x", "f4", (3,)), ("v", "f4", (3,)), ("m23_weight", "f4"),
                ("m23_m_conv", "?"), ("m23_v_conv", "?"), ("ok", "?")]

""" GALAXY_HISTORY_DTYPE gives the propetis of the galaxy at the epoch of
star tagging.
m_star_i - initial M* of the galaxy in M_sun
r_half_2d_i - initial projected 2d half-mass radius (kpc)
r_half_3d_i - initial projected 3D half-mass radius (kpc)
Fe_H_i - initial mass-weighted mean [Fe/H] of stars
sigma_Fe_H_i - the initial 1-sigma scatter in the metallicity distribution
               funciton in dex.
t_50 - the half-mass formation time of stars in Gyr
t_90 - the 90% formation time of stars in Gyr
"""
GALAXY_HISTORY_DTYPE = [("m_star_i", "f4"), ("r_half_2d_i", "f4"),
                        ("r_half_3d_i", "f4"),
                        ("Fe_H_i", "f4"), 
                        ("sigma_Fe_H_i", "f4"),
                        ("a50", "f4"), ("a90", "f4")]

""" STAR_DTYPE gives the properites of an individual star particle.
mp - the stellar mass of the particle in M_sun
Fe_H - [Fe/H] of the particle
t_form - the age of the universe when this star particle formed
"""
STAR_DTYPE = [
    ("mp", "f4"), ("Fe_H", "f4"), ("a_form", "f4")
]

""" PARTICLE_DTYPE is a numpy datatype representing the properties of tracked 
particles within a subhalo. Positions (x) and velocities (v) are given in
physical units.
- x: position in pkpc
- v: velocity in km/s
- snap: integer value for the desired snapshot
 - ok: true if the particle is being tracked and false otherwise
 - smooth: true if the particle smoothly accreted onto the subhalo and false otherwise.
"""
PARTICLE_DTYPE = [
    ("x", "f4", (3,)), ("v", "f4", (3,)),
    ("snap", "i4"), ("id", "i4"),
    ("ok", "?"), ("smooth", "?")
]

""" TREE_COL_NAMES is the mapping of variable names to columns in the
consistent-trees file. These are the variable names you need to pass to
read_tree().
"""
TREE_COL_NAMES = {
    "dfid": 28,
    "id": 1,
    "desc_id": 3,
    "upid": 6,
    "phantom": 8,
    "snap": 31,
    "next_co_prog": 32,
    "mvir": 10,
    "rs": 12,
    "vmax": 16,
    "m200b": 39,
    "m200c": 40,
    "m500c": 41,
    "xoff": 43,
    "spin_bullock": 45,
    "b_to_a": 46,
    "c_to_a": 47,
    "t_to_u": 56,
    "r_vmax": 60,
    "x": 17,
    "v": 20,
    "j": 23,
    "a": 48,
}
_chinchilla_cosmology = { 'flat': True, 'H0': 70.0, 'Om0': 0.286,
                          'Ob0': 0.049, 'sigma8': 0.82, 'ns': 0.95 }
_banerjee_cosmology = { 'flat': True, 'H0': 70.0, 'Om0': 0.3,
                        'Ob0': 0.049, 'sigma8': 0.85, 'ns': 0.95 }
_carmen_cosmology = { 'flat': True, 'H0': 70.0, 'Om0': 0.25,
                      'Ob0': 0.049, 'sigma8': 0.8, 'ns': 1 }

""" parameter_table contains simulation-defining parameters. This includes a 
table of cosmological parameters ("flat", "H0", "Om0", "Ob0", "sigma8", and
"ns"), the particle mass in Msun/h, "mp", the Plummer-equivalent force
softening scale in comoving kpc/h, "eps", and for convience, "h100". "n_snap" is
the number of snapshots in the suite.
"""
parameter_table = {
    "SymphonyLMC": _chinchilla_cosmology.copy(),
    "SymphonyMilkyWay": _chinchilla_cosmology.copy(),
    "EDEN_MilkyWay_8K": _chinchilla_cosmology.copy(),
    "SymphonyMilkyWayDisk": _chinchilla_cosmology.copy(),
    "SymphonyMilkyWayDiskDMO": _chinchilla_cosmology.copy(),
    "SymphonyMilkyWayLR": _chinchilla_cosmology.copy(),
    "SymphonyMilkyWayHR": _chinchilla_cosmology.copy(),
    "SymphonyGroup": _chinchilla_cosmology.copy(),
    "SymphonyLCluster": _banerjee_cosmology.copy(),
    "SymphonyCluster": _carmen_cosmology.copy(),
    "SymphonyClusterCorrupted": _carmen_cosmology.copy(),
    "MWest": _chinchilla_cosmology.copy(),
}

parameter_table["SymphonyLMC"]["eps"] = 0.080
parameter_table["SymphonyMilkyWay"]["eps"] = 0.170
parameter_table["EDEN_MilkyWay_8K"]["eps"] = 0.170
parameter_table["SymphonyMilkyWayDisk"]["eps"] = 0.170*2
parameter_table["SymphonyMilkyWayDiskDMO"]["eps"] = 0.170*2
parameter_table["SymphonyMilkyWayLR"]["eps"] = 0.170
parameter_table["SymphonyMilkyWayHR"]["eps"] = 0.170/2
parameter_table["SymphonyGroup"]["eps"] = 0.360
parameter_table["SymphonyLCluster"]["eps"] = 1.200
parameter_table["SymphonyCluster"]["eps"] = 3.250
parameter_table["SymphonyClusterCorrupted"]["eps"] = 3.250
parameter_table["MWest"]["eps"] = 0.170

parameter_table["SymphonyLMC"]["mp"] = 3.52476e4
parameter_table["SymphonyMilkyWay"]["mp"] = 2.81981e5
parameter_table["EDEN_MilkyWay_8K"]["mp"] = 2.81981e5
parameter_table["SymphonyMilkyWayDisk"]["mp"] = 2.81981e5*8
parameter_table["SymphonyMilkyWayDiskDMO"]["mp"] = 2.81981e5*8
parameter_table["SymphonyMilkyWayLR"]["mp"] = 2.81981e5
parameter_table["SymphonyMilkyWayHR"]["mp"] = 2.81981e5/8
parameter_table["SymphonyGroup"]["mp"] = 2.25585e6
parameter_table["SymphonyLCluster"]["mp"] = 1.51441632e8
parameter_table["SymphonyCluster"]["mp"] = 1.26201360e8
parameter_table["SymphonyClusterCorrupted"]["mp"] = 1.26201360e8
parameter_table["MWest"]["mp"] = 2.81981e5

parameter_table["SymphonyLMC"]["n_snap"] = 236
parameter_table["SymphonyMilkyWay"]["n_snap"] = 236
parameter_table["EDEN_MilkyWay_8K"]["n_snap"] = 236
parameter_table["SymphonyMilkyWayDisk"]["n_snap"] = 236
parameter_table["SymphonyMilkyWayDiskDMO"]["n_snap"] = 236
parameter_table["SymphonyMilkyWayLR"]["n_snap"] = 236
parameter_table["SymphonyMilkyWayHR"]["n_snap"] = 236
parameter_table["SymphonyGroup"]["n_snap"] = 236
parameter_table["SymphonyLCluster"]["n_snap"] = 200
parameter_table["SymphonyCluster"]["n_snap"] = 200
parameter_table["SymphonyClusterCorrupted"]["n_snap"] = 200
parameter_table["MWest"]["n_snap"] = 236

for sim in parameter_table:
    param = parameter_table[sim]
    param["h100"] = param["H0"]/100

parameter_table["ExampleSuite"] = parameter_table["MWest"]

def suite_names():
    """ suite_names returns the names of all the simulation suites.
    """
    return util.SUITE_NAMES

def simulation_parameters(suite_name):
    """ parameter_table contains simulation-defining parameters. This includes a
    table of cosmological parameters ("flat", "H0", "Om0", "Ob0", "sigma8", and
    "ns"), the particle mass in Msun/h, "mp", the Plummer-equivalent force
    softening scale in comoving kpc/h, "eps", and for convience, "h100".
    "n_snap" is the number of snapshots in the suite.
    """
    if suite_name in parameter_table:
        return parameter_table[suite_name]
    else:
        sim_dir = suite_name
        suite_name = halo_dir_to_suite_name(sim_dir)
        if suite_name not in parameter_table:
            raise ValueError("'%s' is neither a recognized suite name nor a simulation direoctry containing a recognized suite name. Recognized suite names are: " + str(suite_names()))
        return parameter_table[suite_name]

def colossus_parameters(param):
    """ colossus_parameters converts a parameter dictionary into a colossus-
    comparitble parameter dictionary.
    """
    var_names = ["Om0", "flat", "H0", "Ob0", "sigma8", "ns"]
    return { name: param[name] for name in var_names }

P_FILE_FMT = "%s/particles/part_%03d.%d"

def halo_dir_to_suite_name(dir_name):
    """ halo_dir_to_suite_name returns the name of the suite that a halo in the
    given directory belongs to.
    """
    if dir_name[-1] == "/": dir_name = dir_name[:-1]
    suite_dir, halo_name = os.path.split(dir_name)
    base_dir, suite_name = os.path.split(suite_dir)
    return suite_name

_SCALE_FACTOR_CACHE = { }

def scale_factors(dir_name):
    """ scale_factors returns the scale factors of each snapshot for a halo in
    the given directory.
    """
    if dir_name in _SCALE_FACTOR_CACHE:
        return _SCALE_FACTOR_CACHE[dir_name]

    # TODO: individual halo-by-halo scale factors
    suite_name = halo_dir_to_suite_name(dir_name)
    if (suite_name in ["SymphonyLMC", "SymphonyGroup", "SymphonyMilkyWayDisk",
                       "SymphonyMilkyWayDiskDMO", "EDEN_MilkyWay_8K",
                      "SymphonyMilkyWay", "MWest", "SymphonyMilkyWayLR",
                       "SymphonyMilkyWayHR"] or
        dir_name in ["SymphonyLMC", "SymphonyGroup", "SymphonyMilkyWayDisk",
                     "SymphonyMilkyWayDiskDMO", "EDEN_MilkyWay_8K",
                     "SymphonyMilkyWay", "MWest", "SymphonyMilkyWayLR",
                     "SymphonyMilkyWayHR"]):
        default = 10**np.linspace(np.log10(0.05), np.log10(1), 236)
    elif (suite_name in ["SymphonyLCluster",  "SymphonyCluster", "SymphonyClusterCorrupted"] or
          dir_name in ["SymphonyLCluster",  "SymphonyCluster", "SymphonyClusterCorrupted"]):
        default = 10**np.linspace(np.log10(0.075), np.log10(1), 200)
    else:
        raise ValueError(("The halo in %s does not belong to a " + 
                          "recognized suite.") % dir_name)

    if dir_name in ["SymphonyLMC", "SymphonyGroup",
                    "SymphonyMilkyWay", "MWest", "SymphonyMilkyWayLR",
                    "SymphonyMilkyWayHR",  "EDEN_MilkyWay_8K",
                    "SymphonyLCluster",  "SymphonyCluster", "SymphonyCluster"]:
        return default

    file_name = path.join(dir_name, "halos", "snap_scale.dat")
    if not path.exists(file_name):
        _SCALE_FACTOR_CACHE[dir_name] = default
        return default

    f = open(file_name, "rb")
    n_snap = struct.unpack("q", f.read(8))[0]
    scale = np.fromfile(f, np.float64, n_snap)
    f.close()

    _SCALE_FACTOR_CACHE[dir_name] = scale
    return scale


def mvir_to_rvir(mvir, a, omega_M):
    """ mvir_to_rvir converts a Bryan & Norman virial mass in Msun/h to a virial
    radius in comoving Mpc/h at a given scale factor a, and omega_M.
    """
    
    if type(mvir) == np.ndarray:
        mvir = np.maximum(np.zeros(len(mvir)), mvir)
    else:
        mvir = max(0, mvir)

    omega_L = 1 - omega_M
    Ez = np.sqrt(omega_M/a**3 + omega_L)
    rho_crit = 2.77519737e11*Ez**2
    omega_Mz = (omega_M/a**3)/Ez**2

    x = omega_Mz - 1
    delta_vir = 18*np.pi**2 + 82*x - 39.0*x**2
    rho_vir = rho_crit*delta_vir

    r_phys = (mvir/(rho_vir * (4*np.pi / 3)))**(1.0/3)
    r_cmov = r_phys/a

    return r_cmov

def flatten(arrays):
    """ flatten takes a list of numpy arrays and flattens it into a single
    array. arrays[i].shape[0] can be any value, but all other components of the
    shape vectors must be the same. THis is needed because hstack doesn't work
    on tensor-arrays.
    """

    N = sum(arr.shape[0] for arr in arrays)

    shape, dtype = arrays[0].shape, arrays[0].dtype
    if len(shape) == 1:
        out = np.zeros(N, dtype=dtype)
    else:
        out = np.zeros((N,) + shape[1:], dtype=dtype)

    start, end = 0, 0
    for i in range(len(arrays)):
        end += arrays[i].shape[0]
        out[start: end] = arrays[i]
        start = end

    return out

def merger_snap(h, x_sub, snap_sub):
    """ merger_snap returns the snapshot where a subhalo with a given set of
    positions and snapshots first falls within h, a halo (i.e. a return value
    of read_mergers()).
    """
    dist = np.sqrt(np.sum((h["x"][snap_sub] - x_sub)**2, axis=1))
    within = dist < h["rvir"][snap_sub]
    merger = h["ok"][snap_sub] & within

    if np.sum(merger) == 0:
        return -1
    else:
        return np.min(snap_sub[merger])

def pristine_merger_indices(b):
    return np.where(b["is_real"] & (~b["is_disappear"]) &
                    b["is_main_sub"] & (b["preprocess"] == -1))[0]

def merger_stats(b, m, x, mvir, snap):
    """ merger_stats returns several useful values: mpeak of the subhalo, the
    scale factor of the merger, the snapshot of the merger, and the mass ratio
    between the host and subhalo at the merger snapshot.
    """
    sub_idx = pristine_merger_indices(b)
    mw = m[0]

    ratio = np.zeros(len(sub_idx))
    m_snap = np.zeros(len(sub_idx), dtype=int)
    mpeak = np.zeros(len(sub_idx))
    
    mw_mass = np.max(mw["mvir"])
    for j, i in enumerate(sub_idx):
        mvir_i = mvir[b["start"][i]: b["end"][i]]
        snap_i = snap[b["start"][i]: b["end"][i]]
        x_i = x[b["start"][i]: b["end"][i]]
                
        m_snap_i = merger_snap(mw, x_i, snap_i)

        m_snap_sub = np.searchsorted(snap_i[::-1], m_snap_i)
        mpeak[j] = np.max(mvir_i)
        m_snap[j] = m_snap_i
        ratio[j] = mvir_i[::-1][m_snap_sub]/mw["mvir"][m_snap_i]

    return mpeak, m_snap, ratio, sub_idx

ROCKSTAR_DTYPE = [("id", "i4"), ("m", "f4"), ("vmax", "f4"), ("rvmax", "f4"),
                 ("x", "f4", (3,)), ("v", "f4", (3,)), ("ok", "?"),
                 ("rvir", "f4"), ("cvir", "f4")]

def read_rockstar(dir_name, comoving=False):
    h, hist = read_subhalos(dir_name, comoving=comoving)
    r = np.zeros(h.shape, dtype=ROCKSTAR_DTYPE)
    r["id"], r["m"], r["vmax"] = h["id"], h["mvir"], h["vmax"]
    r["rvmax"], r["x"], r["v"] = h["rvmax"], h["x"], h["v"]
    r["ok"], r["rvir"], r["cvir"] = h["ok"], h["rvir"], h["cvir"]
    return r, hist

def read_subhalos(dir_name, comoving=False, include_false_selections=False):
    """ read_subhalos reads major merger data from the halo directory dir_name.
    It returns two arrays. The first, m_idx, is the indices of the major
    mergers within the branches arrays. Index 0 is the main halo, index 1 is the
    biggest merger (by Mpeak), index 2 is the second biggest, etc. As a
    convenience, data from the main branches of those haloes are returned as
    the second argument, m. m[halo_idx, snap] gives the  properties of the
    halo at the index halo_idx in m_idx and the snapshot snap. The fields are
    given by SUBHALO_DTYPE (see the header of this file). If a halo doesn't
    exist at a given snapshot, values are set -1 ok will be set to false.

    If comoving is True, the subhaloes will be returned in Rockstar's default,
    comoving units. Otherwise it's returned in symlib's units. If
    include_false_slections is true, haloes which would be included with a
    straight M_peak > 300*mp cutoff are included. If False (the default)
    """
    params = simulation_parameters(dir_name)

    fname = path.join(dir_name, "halos", "subhalos.dat")
    f = open(fname, "rb")

    n_snap = struct.unpack("i", f.read(4))[0]
    n_merger = struct.unpack("i", f.read(4))[0]

    idx = np.fromfile(f, np.int32, n_merger)
    out = np.zeros((n_merger, n_snap), dtype=SUBHALO_DTYPE)
    a = scale_factors(dir_name)

    for i in range(n_merger):
        out["mvir"][i,:] = np.fromfile(f, np.float32, n_snap)
        out["ok"][i,:] = out["mvir"][i,:] > 0
        out["rvir"][i,:] = mvir_to_rvir(out["mvir"][i,:], a, params["Om0"])

    out["rvir"][out["rvir"] == 0] = -1
        
    for i in range(n_merger):
        out["vmax"][i,:] = np.fromfile(f, np.float32, n_snap)
        ok = out["rvir"][i,:] > 0
        out["cvir"][i,ok] = vmax_to_cvir_nfw(
            out["vmax"][i,ok], out["mvir"][i,ok], out["rvir"][i,ok]*a[ok]
        )

    for i in range(n_merger):
        out["rvmax"][i,:] = np.fromfile(f, np.float32, n_snap)
    for i in range(n_merger):
        out["id"][i,:] = np.fromfile(f, np.int32, n_snap)
    for i in range(n_merger):
        out["x"][i,:,:] = np.fromfile(f, (np.float32, (3,)), n_snap)
    for i in range(n_merger):
        out["v"][i,:,:] = np.fromfile(f, (np.float32, (3,)), n_snap)
        
    f.close()

    histories = get_subhalo_histories(out, idx, dir_name)

    file_name = path.join(dir_name, "halos", "core_pos.dat")
    if path.exists(file_name):
        with open(file_name, "rb") as f:
            n_halo, n_snap = struct.unpack("qq", f.read(16))
            x = np.fromfile(f, np.float32, 3*n_halo*n_snap)
            v = np.fromfile(f, np.float32, 3*n_halo*n_snap)
            x = x.reshape((n_halo, n_snap, 3))
            v = v.reshape((n_halo, n_snap, 3))
        
    if not comoving:
        scale = scale_factors(dir_name)
        out = util.set_units_halos(out, scale, params)
        histories = util.set_units_histories(histories, scale, params)

    if not include_false_selections:
        out = out[~histories["false_selection"]]
        histories = histories[~histories["false_selection"]]

    return out, histories

def get_subhalo_histories(s, idx, dir_name):
    if s["x"][0,-1,0] == 0:
        raise ValueError("The input subhalo array has already been " + 
                         "converted to symlib's default units.")

    param = simulation_parameters(dir_name)
    b = read_branches(dir_name)
    h = np.zeros(len(s), dtype=HISTORY_DTYPE)

    # These can just be copied over
    h["start"], h["end"] = b[idx]["start"], b[idx]["end"]
    h["is_real"], h["is_disappear"] = b[idx]["is_real"], b[idx]["is_disappear"]
    h["is_main_sub"] = b[idx]["is_main_sub"]
    # TODO: point into subhalos, not branches
    h["preprocess"] = b[idx]["preprocess"]
    h["first_infall_snap"] = b[idx]["first_infall_snap"]
    h["branch_idx"] = idx

    central = s[0]

    snap = np.arange(len(s[0]))
    
    for i in range(len(s)):
        mvir, vmax, x = s[i]["mvir"], s[i]["vmax"], s[i]["x"]
        ok = mvir > 0
        
        h["mpeak"][i], h["vpeak"][i] = np.max(mvir[ok]), np.max(vmax[ok])
        if i == 0: continue

        m_snap = merger_snap(central, x[ok], snap[ok])
        m_ratio = mvir[m_snap] / central["mvir"][m_snap]

        target = (snap >= m_snap) & s["ok"][i]

        if h["first_infall_snap"][i] == -1: h["first_infall_snap"][i] = m_snap
        h[i]["merger_snap"], h[i]["merger_ratio"] = m_snap, m_ratio
        h["first_infall_snap"][i] = min(h["merger_snap"][i],
                                        h["first_infall_snap"][i])

        infall_snap = h["first_infall_snap"][i]
        mpeak_infall = np.max(s[i,:infall_snap+1]["mvir"])
        h["mpeak_pre"][i] = mpeak_infall
        h["false_selection"][i] = mpeak_infall < 300*param["mp"]

    h["mpeak_pre"][0] = h["mpeak"][0]

    (snap_discrete, snap_eps,
     snap_relax, snap_relax_hydro,
     disrupt_snap, disrupt_snap_rs) = read_convergence_limits(dir_name)
    if snap_discrete is not None:
        ok = ~h["false_selection"]
        h["conv_snap_discrete"][ok] = snap_discrete
        h["conv_snap_eps"][ok] = snap_eps
        h["conv_snap_relax"][ok] = snap_relax
        h["conv_snap_relax_hydro"][ok] = snap_relax_hydro
        h["disrupt_snap"][ok] = disrupt_snap
        h["disrupt_snap_rs"][ok] = disrupt_snap_rs

    return h

def read_convergence_limits(sim_dir):
    file_name = path.join(sim_dir, "convergence_limits.dat")
    if not path.exists(file_name): return None, None, None, None, None, None

    try:
        with open(file_name, "rb") as fp:
            n_halo, = struct.unpack("q", fp.read(8))
            snap_discrete = np.fromfile(fp, np.int32, n_halo)
            snap_eps = np.fromfile(fp, np.int32, n_halo)
            snap_relax = np.fromfile(fp, np.int32, n_halo)
            snap_relax_hydro = np.fromfile(fp, np.int32, n_halo)
            disrupt_snap = np.fromfile(fp, np.int32, n_halo)
            disrupt_snap_rs = np.fromfile(fp, np.int32, n_halo)

            return snap_discrete, snap_eps, snap_relax, snap_relax_hydro, disrupt_snap, disrupt_snap_rs
    except:
        return None, None, None, None, None, None

def read_symfind(dir_name, suffix="fid3"):
    """ read_symfind returns a pair of arrays representing each symfind
    subhalo. dir_name is the simulation's directory. The first is a 2D array
    of type SYMFIND_DTYPE where halos[i,j] is subhalo i at snapshot j.
    Symfind only tracks subhaloes, so there are no entries prioer to first
    infall. The second return value is an array of type HISTORY_DTYPE with
    one element for each subhalo. Subhaloes are ordered by decreasing
    pre-infall Mpeak, and the first element corresponds to the host halo.
    (Because the host is not a subhalo by definition, symfind does not have
    entries for this object).
    """
    h, hist = read_subhalos(dir_name)
    c = read_cores(dir_name)

    s = np.zeros(c.shape, dtype=SYMFIND_DTYPE)
    s["x"], s["v"] = c["x"], c["v"]
    s["r_tidal"], s["r_half"] = c["r_tidal"], c["r50_bound"] 
    s["m_tidal_bound"], s["m"] = c["m_tidal_bound"], c["m_bound"]
    s["m_raw"] = c["m_bound"]
    s["vmax"], s["ok"], s["interp"] = c["vmax"], c["ok"], c["interp"]
    s["f_core"], s["f_core_rs"] = c["f_core"], c["f_core"]
    s["d_core_mbp"] = c["d_core_mbp"]
    s["r_half_rs"], s["ok_rs"] = c["r50_bound_rs"], c["ok_rs"]
    s["is_err"], s["is_err_rs"] = c["is_err"], c["is_err_rs"]

    # Enforce monotonicity to reduce noise in the commonly used "m"
    for i in range(len(s)):
        s["m"][i] = monotonic_m(s["m_raw"][i], s["ok"][i])

    return s, hist

def monotonic_m(m_in, ok):
    m_out = np.zeros(len(m_in))
    idx = np.where(ok)[0]

    prev_max = 0
    for i in idx[::-1]:
        m_out[i] = max(m_in[i], prev_max)
        prev_max = m_out[i]

    return m_out

def read_cores(dir_name, include_false_selections=False, suffix="fid3"):
    """ read_cores read the particle-tracked halo cores of the halo in the
    given directory. The returned array is a structured array of type
    CORE_DTYPE with shape (n_halos, n_snaps).
    """
    if suffix == "":
        file_name = path.join(dir_name, "halos", "cores.dat")
    else:
        file_name = path.join(dir_name, "halos", "cores_%s.dat" % suffix)

    param = simulation_parameters(dir_name)
    mp = param["mp"]/param["h100"]
    h, hist = read_subhalos(dir_name, include_false_selections=True)
    with open(file_name, "rb") as fp:
        n_halo, n_snap = struct.unpack("qq", fp.read(16))
        n = n_halo*n_snap
        out = np.zeros(n, dtype=CORE_DTYPE)

        x = np.fromfile(fp, np.float32, 3*n)
        v = np.fromfile(fp, np.float32, 3*n)

        out["x"] = x.reshape((n,3))
        out["v"] = v.reshape((n,3))
        out["r_tidal"] = np.fromfile(fp, np.float32, n)
        out["r50_bound"] = np.fromfile(fp, np.float32, n)
        out["r50_bound_rs"] = np.fromfile(fp, np.float32, n)
        out["m_tidal"] = np.fromfile(fp, np.float32, n)
        out["m_tidal_bound"] = np.fromfile(fp, np.float32, n)
        out["m_bound"] = np.fromfile(fp, np.float32, n)
        out["vmax"] = np.fromfile(fp, np.float32, n)
        out["f_core"] = np.fromfile(fp, np.float32, n)
        out["f_core_rs"] = np.fromfile(fp, np.float32, n)
        out["d_core_mbp"] = np.fromfile(fp, np.float32, n)
        out["ok"] = np.fromfile(fp, np.bool_, n)
        out["ok_rs"] = np.fromfile(fp, np.bool_, n)
        out["is_err"] = np.fromfile(fp, np.bool_, n)
        out["is_err_rs"] = np.fromfile(fp, np.bool_, n)
        out["interp"] = np.fromfile(fp, np.bool_, n)
       

    out = out.reshape((n_halo, n_snap))

    if not include_false_selections:
        h, hist = read_subhalos(dir_name, include_false_selections=True)
        out = out[~hist["false_selection"]]

    return out

def read_um(dir_name):
    fname = path.join(dir_name, "um", "um.dat")
    h, hist = read_subhalos(dir_name)
    h_cmov, hist_cmov = read_subhalos(dir_name, comoving=True)
    n_halo, n_snap = h.shape
    n = n_snap*n_halo

    out = np.zeros(n, dtype=UM_DTYPE)

    with open(fname) as fp:
        out["m_star"] = np.fromfile(fp, np.float32, n)
        out["m_in_situ"] = np.fromfile(fp, np.float32, n)
        x = np.fromfile(fp, np.float32, 3*n)
        v = np.fromfile(fp, np.float32, 3*n)
        out["x"] = x.reshape((n,3))
        out["v"] = v.reshape((n,3))
        out["sfr"] = np.fromfile(fp, np.float32, n)
        out["rank"] = np.fromfile(fp, np.float32, n)
        out["mvir"] = np.fromfile(fp, np.float32, n)
        out["vmax"] = np.fromfile(fp, np.float32, n)
        out["is_orphan"] = np.fromfile(fp, np.bool, n)
        out["ok"] = np.fromfile(fp, np.bool, n)
        
    out["x"] *= 1e3

    out = out.reshape((n_halo, n_snap))
    out["is_orphan"] = out["ok"] & (~h["ok"])
    return out

def read_galaxies(dir_name, model="um"):
    fname = path.join(dir_name, "galaxies", "%s.dat" % model)

    h, _ = read_rockstar(dir_name)
    n_halo, n_snap = h.shape
    n = n_snap*n_halo

    gal = np.zeros(n, dtype=GALAXY_DTYPE)
    gal_hist = np.zeros(n_halo, dtype=GALAXY_HISTORY_DTYPE)

    with open(fname) as fp:
        gal_hist["m_star_i"] = np.fromfile(fp, np.float32, n_halo)
        gal_hist["r_half_i"] = np.fromfile(fp, np.float32, n_halo)

        gal["m_star"] = np.fromfile(fp, np.float32, n)
        gal["r_half"] = np.fromfile(fp, np.float32, n)
        gal["m_dyn"] = np.fromfile(fp, np.float32, n)
        x = np.fromfile(fp, np.float32, 3*n)
        v = np.fromfile(fp, np.float32, 3*n)
        gal["x"] = x.reshape((n,3))
        gal["v"] = v.reshape((n,3))
        gal["v_disp_3d_dm"] = np.fromfile(fp, np.float32, n)
        gal["v_disp_3d_star"] = np.fromfile(fp, np.float32, n)
        gal["vmax_dm"] = np.fromfile(fp, np.float32, n)
        gal["vmax_dm_debias"] = np.fromfile(fp, np.float32, n)
        gal["m23_weight"] = np.fromfile(fp, np.float32, n)
        gal["m23_m_conv"] = np.fromfile(fp, bool, n)
        gal["m23_v_conv"] = np.fromfile(fp, bool, n)
        gal["ok"] = np.fromfile(fp, np.bool, n)
        
    gal = gal.reshape((n_halo, n_snap))
    
    return gal, gal_hist

def read_branches(dir_name):
    """ read_branches reads main branch data from the halo directory dir_name.
    It returns an array with length n_branches where each element has type
    BRANCH_DTYPE.
    """
    fname = path.join(dir_name, "halos", "tree_header.dat")
    f = open(fname, "rb")
    
    n = struct.unpack("i", f.read(4))[0]
    central_idx = struct.unpack("i", f.read(4))[0]
    out = np.zeros(n, dtype=BRANCH_DTYPE)

    edges = np.fromfile(f, np.int32, n+1)
    out["start"] = edges[:-1]
    out["end"] = edges[1:]
    out["is_real"] = np.fromfile(f, bool, n)
    out["is_disappear"] = np.fromfile(f, bool, n)
    out["is_main_sub"] = np.fromfile(f, bool, n)
    out["preprocess"] = np.fromfile(f, np.int32, n)
    out["first_infall_snap"] = np.fromfile(f, np.int32, n)

    f.close()
    return out    

def read_tree(dir_name, var_names):
    """ read_tree reads variables from the halo directory halo_dir in
    depth-first order. var_names is a list of variables to be read (see
    TREE_COL_NAMES). A list of arrays is returned. Use the branches and merger
    files to identify main branches and important haloes, respectively.
    """
    tree_files = sorted(glob.glob(path.join(dir_name, "trees", "tree_*.dat")))
    tree_files = [fname for fname in tree_files if
                  path.basename(fname) != "tree_header.dat"]

    out = []
    for i in range(len(var_names)):
        var = []
        for j in range(len(tree_files)):
            hd = read_tree_header(tree_files[j])
            offset = tree_var_offset(hd, var_names[i])
            f = open(tree_files[j], "rb")
            col = tree_var_col(hd, var_names[i])
            f.seek(offset)
        
            if is_int(col, hd):
                var.append(np.fromfile(f, np.int32, hd.n))                
            elif is_float(col, hd):
                var.append(np.fromfile(f, np.float32, hd.n))
            else:
                var.append(np.fromfile(f, (np.float32, (3,)), hd.n))

        out.append(flatten(var))
    return out

def propagate_parent_indices(idx):
    """ propagate_parent_indices performs a union-find on the array idx so that
    if subhalo A was preprocessed by B which was preprocessed by C, A's index
    points to C.

    This is currently brute-force and very slow. There's a well-known union-find
    algorithm which is faster. Can switch to this if it ever matters, but it's
    also ~130 lines in my old Julia implementation, so I'm not sure that it's 
    worht it.
    """
    n_changed = -1
    while n_changed != 0:
        n_changed = 0
        for i in range(len(idx)):
            if idx[i] == i: idx[i] = -1
            if idx[i] != -1 and idx[[idx[i]]] != -1:
                n_changed += 1
                idx[i] = idx[idx[i]]

def pre_infall_host(hist, first_order=False):
    if first_order:
        pre = hist["preprocess"]
        idx = np.arange(len(hist), dtype=int)
        order = np.argsort(hist["branch_idx"])
        idx, branch = idx[order], hist["branch_idx"][order]
        
        i = np.searchsorted(branch, pre)
        i[i > len(i) - 1] = len(i) - 1

        ext_host = pre != branch[i]
        no_host = hist["preprocess"] == -1

        host = idx[i]
        host[ext_host] = 0
        host[no_host] = -1

        return host
    else:
        host = pre_infall_host(hist, first_order=True)

        n_changed = -1
        while n_changed != 0:
            n_changed = 0
            for i in range(len(host)):
                if host[i] > 0 and host[host[i]] > 0:
                    n_changed += 1
                    host[i] = host[host[i]]
        return host


def read_merger_idxs(dir_name):
    halo_name = dir_name.split("/")[-1]
    parent_dir = "/".join(dir_name.split("/")[:-1])
    merger_idxs_name = path.join(parent_dir, "merger_idxs.txt")
    lmc_idx_cosmo, lmc_idx_zoom, gse_idx = np.loadtxt(
        merger_idxs_name, usecols=(1, 4, 7), dtype=int).T
    halo_names = np.loadtxt(merger_idxs_name, usecols=(0,), dtype=str)
    for i in range(len(halo_names)):
        if halo_name == halo_names[i]:
            return lmc_idx_cosmo[i], lmc_idx_zoom[i], gse_idx[i]
    raise ValueError("Could not find %s in %s" % (halo_name, merger_idxs_name))

# Everything else in this file is an internal helper function

def is_int(col, hd): return col < hd.n_int
def is_float(col, hd): return col >= hd.n_int and col < hd.n_float
     
def read_tree_header(fname):
    f = open(fname, "rb")
    class Header(object): pass
    hd = Header()
    hd.n, hd.n_int, hd.n_float, hd.n_vec = struct.unpack("iiii", f.read(16))
    cols = array.array("i")
    cols.fromfile(f, hd.n)
    hd.cols = np.asarray(cols, dtype=np.int32)
    return hd

def tree_var_col(hd, var_name):
    if var_name not in TREE_COL_NAMES:
        raise ValueError("Variable name '%s' unrecognized" % var_name)
    
    target_col = TREE_COL_NAMES[var_name]
    for i in range(len(hd.cols)):
        if hd.cols[i] == target_col: break

    return i

def tree_var_offset(hd, var_name):
    i = tree_var_col(hd, var_name)
    hd_size = 4*4 + 4*(hd.n_int + hd.n_float + hd.n_vec)
    return hd.n*4*(i + 2*max(0, i - hd.n_int - hd.n_float)) + hd_size    

def merger_lookup_table(b, dfid):
    """ merger_lookup_table creates a look-up table for finding the branches of
    mergers. It's used along side find_merger_branch and
    find_all_merger_branches.

    b is an array of branches (BRANCH_DTYPE) and dfid is an array of
    depth-first IDs with merger tree ordering (can be read by calling 
    read_tree on "dfid")
    """
    return dfid[b["start"]]

def find_merger_branch(lookup_table, co_prog):
    """ find_merger_branch returns the index of the branch corresponding to
    a given co-progenitor ID.
    """
    return np.searchsorted(lookup_table, co_prog)

def find_all_merger_branches(b, lookup_table, co_prog, i):
    """ find_all_merger_branches returns an array with the indexes of all the
    branches that merge with a halo in the given snapshot (i.e. this is the
    last resolved snapshot of those branches). b is an array of branches,
    (BRANCH_DTYPE), lookup_table is a table returned by merger_lookup_table,
    co_prog is an array of co-progenitor IDs (can be read by calling read_tree
    on "next_co_prog"), and i is the index of the halo whose mergers you're
    trying to find within the merger tree.
    """
    branches = []
    while co_prog[i] != -1:
        bi = find_merger_branch(lookup_table, co_prog[i])
        branches.append(bi)
        i = b["start"][bi]

    return np.array(branches, dtype=int)
        
def read_particle_header(base_dir):
    n_snap = struct.unpack("i", f.read(4))[0]
    n_merger = struct.unpack("i", f.read(4))[0]

    idx = np.fromfile(f, np.int32, n_merger)    
    
class ParticleHeader(object):
    def __init__(self, base_dir):
        file_name = path.join(base_dir, "particles", "particle_header.dat")
        try:
            f = open(file_name, "rb")
        except FileNotFoundError:
            raise FileNotFoundError("The particle data for this halo does not exist.") from None
        # TODO: if not working, try os.path.exists() and os.path.isfile() after that

        self.n_file = struct.unpack("i", f.read(4))[0]
        self.n_halo = struct.unpack("i", f.read(4))[0]
        self.n_particle = struct.unpack("i", f.read(4))[0]
        
        self.file_lengths = np.fromfile(f, np.int32, self.n_file)
        self.offsets = np.fromfile(f, np.int32, self.n_halo)
        self.sizes = np.fromfile(f, np.int32, self.n_halo)
        self.file_idxs = np.fromfile(f, np.int32, self.n_halo)
        self.n0 = np.fromfile(f, np.int32, self.n_halo)

        self.base_dir = base_dir
        
        f.close()

class ParticleTags(object):
    def __init__(self, base_dir, part_hd):
        self.id = [None]*part_hd.n_halo
        self.snap = [None]*part_hd.n_halo
        self.flag = [None]*part_hd.n_halo

        for i_file in range(part_hd.n_file):
            file_name = path.join(base_dir, "particles",
                                  "tags.%d.dat" % i_file)
            f = open(file_name, "rb")
            
            to_read = np.where(i_file == part_hd.file_idxs)[0]
            for i in to_read:
                self.id[i] = np.fromfile(f, np.int32, part_hd.sizes[i])
                self.snap[i] = np.fromfile(f, np.int16, part_hd.sizes[i])
                self.flag[i] = np.fromfile(f, np.uint8, part_hd.sizes[i])

            f.close()

        self.n0 = part_hd.n0
        self.n_file = part_hd.n_file
        self.n_halo =  part_hd.n_halo
        self.file_idxs = part_hd.file_idxs

        self._n0_snaps = [self.snap[i][self.flag[i] == 0]
                          for i in range(part_hd.n_halo)]
        self._offset_lookup = {}

    def halo_offset(self, snap, i_halo):
        if snap in self._offset_lookup:
            return self._offset_lookup[snap][i_halo]
        else:
            offsets = np.zeros(self.n_halo, dtype=int)
            for i_file in range(self.n_file):
                to_read = np.where(i_file == self.file_idxs)[0]
                curr_offset = 4
                for i in to_read:
                    offsets[i] = curr_offset

                    if len(self._n0_snaps[i]) == 0: continue
                    n = np.sum(self._n0_snaps[i] <= snap)
                    if n == 0: continue

                    curr_offset += 6*n + 8 + 12 + 12
            self._offset_lookup[snap] = offsets
            return self._offset_lookup[snap][i_halo]

class TagLookup(object):
    def __init__(self, base_dir):
        file_name = path.join(base_dir, "particles", "tags.lookup_table.dat")
        f = open(file_name, "rb")
        Np = struct.unpack("i", f.read(4))[0]
        self.halo = np.fromfile(f, np.int16, Np)
        self.index = np.fromfile(f, np.int32, Np)
        
class ParticleInfo(object):
    def __init__(self, base_dir):
        self.part_hd = ParticleHeader(base_dir)
        self.tags = ParticleTags(base_dir, self.part_hd)
        self.lookup = TagLookup(base_dir)

        global_offset = np.zeros(self.part_hd.n_halo, dtype=int)
        global_offset[1:] = np.cumsum(self.part_hd.n0[:-1])
        self.global_offset = global_offset
        self.global_index = global_offset[self.lookup.halo] + self.lookup.index
        
        # Used to find false selections
        self.h_false, self.hist_false = read_subhalos(
            base_dir, include_false_selections=True)

def is_real_confirmed(part_info, h, i_sub):
    first_snap = np.where(h[i_sub]["ok"])[0][0]
    ok = read_particles(part_info, None, first_snap, "valid", owner=i_sub)
    return np.sum(ok) > 0

class Particles(object):
    """ A wrapper for accessing particle data associated with the desired halo suite."""
    def __init__(self, sim_dir):
        self.sim_dir = sim_dir
        self.part_info = ParticleInfo(sim_dir)
        self.params = simulation_parameters(sim_dir)
        self.scale = scale_factors(sim_dir)
        self.h_cmov, _ = read_subhalos(sim_dir, comoving=True)

    def read(self, snap, halo=-1, mode="current", comoving=False):
        """ read returns a list of 1D arrays containing frequently used
        variables stored in the particle data for a given subhalo. 

        Parameters
        ----------
        snap: int, snapshot.
        halo: int, index for subhalo of interest, default returns all subhalos.
        comoving: boolean, default is False to return physical units.
        mode: string 
            "current" is the default setting and returns variables for
            valid particles.
            "all" returns variables for all particles, valid or invalid.
            "smooth" returns variables for smoothly accreted particles. 
            "stars" returns variables for particles that can host stars. (this
            is the same as "smooth" mode)

        Returns
        -------
        p: list containing 1D arrays of the following variables from stored
        particle data: "id", "x", "v", "snap", "ok", and "smooth". See
        PARTICLE_DTYPE for more detail.
        """

        sim_dir = self.sim_dir
        part_info = self.part_info
        h0 = self.h_cmov[0,snap]
        a = self.scale[snap]

        # These modes need to read in all particles simultaneously: cannot
        # just read one halo.
        if mode == "all" or mode == "current":
            idp = read_particles(part_info, sim_dir, snap, "id")
            x = read_particles(part_info, sim_dir, snap, "x")
            v = read_particles(part_info, sim_dir, snap, "v")
            if not comoving:
                for s in range(len(x)):
                    x[s] = util.set_units_x(x[s], h0, a, self.params)
                    v[s] = util.set_units_v(v[s], h0, a, self.params)
            
            snaps = read_particles(part_info, sim_dir, snap, "snap")
            valid = read_particles(part_info, sim_dir, snap, "valid")
            smooth = read_particles(part_info, sim_dir, snap, "ownership")
            
            # Remove invalid particles.
            if mode == "current":
                for i in range(len(x)):
                    ok = valid[i]
                    x[i], v[i], idp[i] = x[i][ok], v[i][ok], idp[i][ok]
                    snaps[i], smooth[i] = snaps[i][ok], smooth[i][ok]
                    valid[i] = valid[i][ok]

        if mode == "smooth" or mode == "stars":
            idp = [None]*len(self.h_cmov)
            x = [None]*len(self.h_cmov)
            v = [None]*len(self.h_cmov)
            snaps = [None]*len(self.h_cmov)
            valid = [None]*len(self.h_cmov)
            smooth = [None]*len(self.h_cmov)
            for sh in range(len(self.h_cmov)):
                # A single halo read is fast for smooth particles, so we only
                # have to read the specified halo in this mode
                if halo != -1 and sh != halo: continue
                idp[sh] = read_particles(part_info, sim_dir, snap, "id", owner=sh)
                x[sh] = read_particles(part_info, sim_dir, snap, "x", owner=sh)
                v[sh] = read_particles(part_info, sim_dir, snap, "v", owner=sh)
                if not comoving:
                    x[sh] = util.set_units_x(x[sh], h0, a, self.params)
                    v[sh] = util.set_units_v(v[sh], h0, a, self.params)
            
                snaps[sh] = read_particles(part_info, sim_dir, snap,
                                           "snap", owner=sh)
                valid[sh] = read_particles(part_info, sim_dir, snap,
                                           "valid", owner=sh)
                smooth[sh] = read_particles(part_info, sim_dir, snap,
                                            "ownership", owner=sh)
        p = [None]*len(x)
        for i in range(len(x)):
            if halo == -1 or i == halo:
                p[i] = np.zeros(len(x[i]), dtype=PARTICLE_DTYPE)
                p[i]["id"] = idp[i]
                p[i]["x"] = x[i]
                p[i]["v"] = v[i]
                p[i]["snap"] = snaps[i]
                p[i]["ok"] = valid[i]
                p[i]["smooth"] = smooth[i] == 0

        if halo == -1:
            return p 
        else:
            return p[halo]

    def core_indices(self, snap=-1, halo=-1, mode="current"):
        """ core_indices returns the indices of the "core" particles for
        subhalos. Core particles are the 32 most-bound smoothly-accreted
        particles at the snapshot of first infall for the subhalo. These
        particles are ordered by boundedness during this snapshot and will be
        set to -1 if that core particle is invalid (either due to the indexing
        mode or because the subhalo had fewer than 32 particles).

        The indices of these particles change depending on which mode of
        particle array they are indexing into. This can be set with the mode
        parameter and has the same meaning as the read method.

        The host halo does not have any core particles.

        Parameters
        ----------
        snap: int, snapshot. Only needed for the "current" mode.
        halo: int, index for subhalo of interest, default returns all subhalos.
        mode: string 
            "current" is the default and indexes across all currently-accreted
            particles. Indices will be negative if the particle hasn't been
            accreted before the chosen snapshot.
            "all" indexes across all particles ever accreted by the subhalo
            "smooth" indexes over 

        Returns
        -------
        idx: A (N_subhalo, 32)-shape array giving core particle indices of
           each subhalo. If the halo parameter is set to a non-negative
           integer, only the row for that halo will be returned.
        """
        idx = read_particles(self.part_info, self.sim_dir,
                             0, "infall_core")
        idx = np.array(idx)
        if mode == "all":
            pass
        elif mode == "smooth" or mode == "stars":
            smooth = read_particles(self.part_info, self.sim_dir,
                                    0, "ownership")
            for sh in range(len(smooth)):
                if halo == -1 or sh == halo:
                    smooth_i = smooth[sh]==0
                    idx_smooth = np.cumsum(smooth_i)-1
                    idx[sh] = idx_smooth[idx[sh]]
        elif mode == "current":
            if snap == -1:
                raise ValueError("The snap variable must be set when calling core_indices in mode 'current'")
            valid = read_particles(self.part_info, self.sim_dir, snap, "valid")
            for sh in range(len(valid)):
                if halo == -1 or sh == halo:
                    valid_i = np.asarray(valid[sh], dtype=int)
                    idx_valid = np.cumsum(valid_i)-1
                    idx[sh] = idx_valid[idx[sh]]
                    is_ok = np.asarray(valid[sh][idx[sh]], dtype=bool)
                    idx[sh,~is_ok] = -1
        else:
            raise ValueError("Unrecognized mode '%s'" % mode)

        if halo == -1:
            idx[0] = -1
            return idx
        else:
            return idx[halo]

    def count(self, snap=-1, halo=-1, mode="current"):
        """ core_indices returns the number of particles in each subhalo
        according to different read modes.

        Parameters
        ----------
        snap: int, snapshot. Only needed for the "current" mode.
        halo: int, index for subhalo of interest, default returns all subhalos.
        mode: string 
            "current" is the default and gives the number of all
            currently-accreted particles.
            "all" returns the number of accreted particles by z=0
            "smooth" returns the number of smoothly accreted particles by z=0

        Returns
        -------
        idx: A (N_subhalo, 32)-shape array giving core particle indices of
           each subhalo. If the halo parameter is set to a non-negative
           integer, only the row for that halo will be returned.
        """
        # Despite the internal names, this doesn't require actually doing any
        # reads. These arrays are already loaded into RAM.
        if mode == "current":
            valid = read_particles(self.part_info, self.sim_dir, snap, "valid")
            count = np.zeros(len(valid), dtype=int)
            for sh in range(len(smooth)):
                if halo == -1 or halo == sh:
                    count[sh] = np.sum(valid[sh])
        else:
            smooth = read_particles(self.part_info, self.sim_dir,
                                    snap, "ownership")
            count = np.zeros(len(smooth), dtype=int)
            for sh in range(len(smooth)):
                if halo == -1 or halo == sh:
                    if mode == "smooth" or mode == "stars":
                        count[sh] = np.sum(smooth[sh] == 0)
                    elif mode == "all":
                        count[sh] = len(smooth[sh])
                    else:
                        raise ValueError("Unrecoganized mode '%s'" % mode)

        if halo == -1:
            return count
        else:
            return count[halo]


def read_particles(part_info, base_dir, snap, var_name,
                   owner=None, include_false_selections=False):
    hd, tags = part_info.part_hd, part_info.tags
    
    h_idx = np.arange(len(part_info.hist_false), dtype=int)
    if include_false_selections:
        false_remapping = h_idx
    else:
        false_remapping = h_idx[~part_info.hist_false["false_selection"]]
    if owner is not None and not include_false_selections:
        owner = false_remapping[owner]

    if var_name == "id":
        if owner is None:
            return [tags.id[i] for i in false_remapping]
        else:
            return tags.id[owner][tags.flag[owner] == 0]
    elif var_name == "snap":
        if owner is None:
            return [tags.snap[i] for i in false_remapping]
        else:
            return tags.snap[owner][tags.flag[owner] == 0]
    elif var_name == "ownership":
        if owner is None:
            return [tags.flag[i] for i in false_remapping]
        else:
            return tags.flag[owner][tags.flag[owner] == 0]
    elif var_name == "valid":
        if owner is None:
            valid = [None]*len(tags.snap)
            for i in range(len(valid)):
                valid[i] = tags.snap[i] <= snap
            return [valid[i] for i in false_remapping]
        else:
            valid = (tags.snap[owner] <= snap)[tags.flag[owner] == 0]
            return valid
    elif var_name in ["x", "v"]:
        snap_name = "snap_%03d" % snap
        if owner is not None:
            first_snap = np.min(np.where(part_info.h_false["ok"][owner,:])[0])
            if snap < first_snap:
                num = int(np.sum(tags.flag[owner] == 0))
                return np.ones((num, 3)) * np.nan

            i_file = hd.file_idxs[owner]
            file_name = "%s.%d.dat" % (var_name, i_file)

            path = os.path.join(hd.base_dir, "particles", snap_name, file_name)
            f = open(path, "rb")

            code = struct.unpack("i", f.read(4))[0]
            if code != 0: raise ValueError("%s is not a vector file." % path)

            offset = tags.halo_offset(snap, owner)
            f.seek(offset, 0)
            
            size = struct.unpack("q", f.read(8))[0]
            min = np.array(struct.unpack("fff", f.read(12)))
            max = np.array(struct.unpack("fff", f.read(12)))
            
            qx_i = np.fromfile(f, np.uint16, size*3)
            x_i = _dequantize_vector(qx_i, min, max)
            x_i = _expand_vector(tags, x_i, owner, snap)

            f.close()
            return x_i

        x_full = np.zeros((hd.n_particle, 3))

        for i_file in range(hd.n_file):
            file_name = "%s.%d.dat" % (var_name, i_file)
            path = os.path.join(hd.base_dir, "particles", snap_name, file_name)
            f = open(path, "rb")
            
            code = struct.unpack("i", f.read(4))[0]
            if code != 0: raise ValueError("%s is not a vector file." % path)

            to_read = np.where(hd.file_idxs == i_file)[0]
            for i_halo in to_read:
                if hd.n0[i_halo] == 0: continue
                ok = tags.snap[i_halo][tags.flag[i_halo] == 0] <= snap
                if np.sum(ok) == 0: continue

                curr_offset = f.seek(0, 1)

                size = struct.unpack("q", f.read(8))[0]
                min = np.array(struct.unpack("fff", f.read(12)))
                max = np.array(struct.unpack("fff", f.read(12)))
                
                qx_i = np.fromfile(f, np.uint16, size*3)
                x_i = _dequantize_vector(qx_i, min, max)

                x_i = _expand_vector(tags, x_i, i_halo, snap)
                offset = part_info.global_offset
                x_full[offset[i_halo]: offset[i_halo]+len(x_i)] = x_i

            f.close()
            
        out = [None]*hd.n_halo
        for i_halo in range(hd.n_halo):
            idx = part_info.global_index[tags.id[i_halo] - 1]
            out[i_halo] = x_full[idx]

        return [out[i] for i in false_remapping]
    elif var_name == "infall_core":
        # Setting owner doesn't do anything for infall_core.

        file_name = os.path.join(base_dir, "halos", "infall_cores.dat")
    
        with open(file_name, "rb") as fp:
            n_halo, n_core = struct.unpack("qq", fp.read(16))
            idxs = np.fromfile(fp, dtype=np.int32, count=n_halo*n_core)
            idxs = idxs.reshape((n_halo, n_core))
        return [idxs[i] for i in false_remapping]
    else:
        raise ValueError("Unknown property name, '%s'" % var_name)

def transform_smooth_particles(x, part_info, mode="all", ok=None):
    hd, tags = part_info.part_hd, part_info.tags
    if len(x) != hd.n_halo:
        raise ValueError("part_info (n_halo = %d) does not match given set of particle data (n_halo = %d)" % (hd.n_halo, len(x)))

    if mode == "smooth" or mode == "stars": return x

    if len(x[0].shape) == 1:
        full_shape = (hd.n_particle,)
    else:
        full_shape = (hd.n_particle,) + x[0].shape[1:]
    
    x_full = np.zeros(full_shape, dtype=x[0].dtype)
    
    offset = part_info.global_offset
    for i_halo in range(hd.n_halo):
        x_i = x[i_halo]
        x_full[offset[i_halo]: offset[i_halo]+len(x_i)] = x_i
            
    out = [None]*hd.n_halo
    for i_halo in range(hd.n_halo):
        idx = part_info.global_index[tags.id[i_halo] - 1]
        out[i_halo] = x_full[idx]

    if mode == "all": return out

    if mode != "current": raise ValueError("Unknown mode '%s'" % mode)
    if ok is None:
        raise ValueError("Must pass an ok list to use mode='current'")

    for i_halo in range(len(out)):
        out[i_halo] = out[i_halo][ok[i_halo]]

    return out

def _dequantize_vector(qx, min, max):
    dx = max - min
    n = len(qx)//3
    
    qx_vec = qx.reshape((n, 3))
    uint16_max = float(np.iinfo(np.uint16).max)
    out = np.zeros((n, 3))
    for dim in range(3):
        out[:,dim] = dx[dim]/uint16_max*(qx_vec[:,dim]+ random.random(n)) + min[dim]

    return out
         
def _expand_vector(tags, x, i, snap):
    ok = tags.snap[i][tags.flag[i] == 0] <= snap
    out = np.ones((len(ok), 3)) * np.nan
    if np.sum(ok) != len(x):
        print("   ", np.sum(ok), len(x))
    assert(np.sum(ok) == len(x))
    out[ok] = x
    return out
    
# NFW math
def _two_way_interpolate(x, y):
    """ two_way_interpolate returns interpolation functions that map from x -> y
    and y -> x. Both input arrays must monotonic.
    """
    if x[0] < x[1]:
        x_to_y = interpolate.interp1d(x, y)
    else:
        x_to_y = interpolate.interp1d(x[::-1], y[::-1])

    if y[0] < y[1]:
        y_to_x = interpolate.interp1d(y, x)
    else:
        y_to_x = interpolate.interp1d(y[::-1], x[::-1])

    return x_to_y, y_to_x

def _f(x): return np.log(1+x) - x/(1+x)
def _x_max_nfw(): return 2.1626
def _v_vmax_nfw(x): return 2.1506 * np.sqrt(_f(x) / x)
def _m_enc_nfw(x): return _f(x)
def _alpha_nfw(x): return -1 - 2*x/(1 + x)

def _half_mass_nfw(x, mass_fraction):
    def f_mass_ratio(xx):
        return _m_enc_nfw(xx) / _m_enc_nfw(x) - mass_fraction
    sol = optimize.root_scalar(f_mass_ratio, bracket=[1e-4*x, x])
    return sol.root

_c = 10**np.linspace(0, 3, 1000)
_cv = np.sqrt(_f(_x_max_nfw()) / _x_max_nfw() * _c/_f(_c))
# Make sure we're solving the correct part of the cV - cvir relation
_cv[_c < _x_max_nfw()] = 1.0 
c_to_cv_nfw, cv_to_c_nfw = _two_way_interpolate(_c, _cv)

def vmax_to_cvir_nfw(vmax, mvir, rvir):
    vvir = 655.8 * (mvir/1e14)**0.5 * (rvir/1.0)**-0.5
    cv = vmax/vvir
    cv[cv < 1] = 1
    return cv_to_c_nfw(cv)    

def main():
    base_dir = "/oak/stanford/orgs/kipac/users/phil1/simulations/ZoomIns"
    suite = "SymphonyMilkyWay"
    i_halo = 0
    sim_dir = get_host_directory(base_dir, suite, i_halo)

    param = simulation_parameters(suite)
    h, hist = read_subhalos(param, sim_dir)
    
    info = ParticleInfo(sim_dir)

    x = read_particles(info, sim_dir, 235, "x")
    v = read_particles(info, sim_dir, 235, "v")
    ok = read_particles(info, sim_dir, 235, "valid")
            
if __name__ == "__main__": main()



import numpy as np
import matplotlib.pyplot as plt
import numpy.random as random
import scipy.stats as stats
import abc
import scipy.optimize as optimize
import scipy.integrate as integrate
import scipy.interpolate as interpolate
from . import lib
from . import util
from . import sampling
import time

# cosmology information
from colossus.cosmology import cosmology
from colossus.halo import mass_so

""" 
Variables names:

'mpeak' - Peak Brayn & Norman virial mass (Msun)
'mvir' - Bryan & Norman virial mass (Msun)
'rvir' - Bryan & Norman virial radius (pkpc)
'cvir' - NFW concentration relative to the Bryan & Norman virial radius
'z' redshift
'mstar' - The stellar mass assinged to this halo (Msun)
"""

NIL_RANK = -1

DEFAULT_QUANTILE_EDGES = np.zeros(13)
DEFAULT_QUANTILE_EDGES[1:] = 10**np.linspace(-4, 0, 12)

DEFAULT_R_BINS = np.zeros(102)
DEFAULT_R_BINS[1:] = 10**np.linspace(-2.5, 0, 101)

MIN_PARTICLES_PER_QUANTILE = 10

DEFAULT_CORE_PARTICLES = 30

#########################
# Abstract base classes #
#########################

class ProfileShapeModel(abc.ABC):
    """ AbstractProfile is an abstract base class for galaxy profile models. It
    allows you to convert a half-light radius into an enclosed stellar mass
    profile.
    """
    
    @abc.abstractmethod
    def m_enc(self, m_star, r_half, r, **kwargs):
        """ m_enc returns the enclosed mass profile as a function of 3D radius,
        r. m_star is the asymptotic stellar mass of the galaxy, r_half is the 2D
        half-light radius of the galaxy. Returned masses will be in the same
        units as m_star.
        """
        pass
        
    @abc.abstractmethod
    def density(self, m_star, r_half, r, **kwargs):
        """ density returns the local density as a function of 3D radius, r.
        m_star is the asymptotic stellas mass of the galaxy, r_hald is the 2D
        half-light radius of the galaxy. Returned masses will be in the same
        units of m_star.
        """
        pass

    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out


class RHalfModel(abc.ABC):
    """ RHalfModel is an abstract base class for models of galaxy half-mass
    radii. 
    """
    
    @abc.abstractmethod
    def r_half(self, **kwargs):
        """ r_half returns the half-mass radius of a a galaxy in physical kpc.
        """
        pass

    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out
    
class MStarModel(abc.ABC):
    """ MStarModel is an abstract base class for models of the Mhalo-Mstar
    relation.
    """

    @abc.abstractmethod
    def m_star(self, **kwargs):
        """ m_star returns the stellar mass of a galaxy in Msun.
        """
        pass

    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out

class SFHModel(abc.ABC):
    """ SFHModel is an abstract base class for models that compute star
    formation histories.
    """
    def sfh(self, **kwargs):
        """ sfh returns a (2, n_snap) array. sfh[0,:] = time, sf[1,:] = M*.
        """
        pass

    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out


class FeHMeanModel(abc.ABC):
    """ MeanFeHModel is an abstract base class for models of the
    mean [Fe/H] values of galaxies
    """
    @abc.abstractmethod
    def Fe_H(self, **kwargs):
        """ Fe_H returns the mean [Fe/H] of the galaxy
        """
        pass

    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out
    
class FeHMDFModel(abc.ABC):
    """ FeHMDFModel is an abstract base class for models of the metallicitiy
    distirbution funciton of [Fe/H]. 
    """
    @abc.abstractmethod
    def mdf(self, Fe_H_mean, **kwargs):
        """ mdf returns an unnormalized  1d funtion defined on the range
        [-10, 5] which gives the pdf of metallicities in the galaxy, given 
        a target mean metallicity.
        """
        pass

    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out

class FeHProfileModel(abc.ABC):
    """ FeHProfileModel is an abstract base class for models of the metallicitiy
    distirbution funciton of [Fe/H]. 
    """
    @abc.abstractmethod
    def Fe_H_profile(self, FeH_MDF, ranks, **kwargs):
        """ FeH_profile returns Fe/H values for each star particle based
        on a given Fe/H MDF
        """
        pass
        
    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out

class MetalCorrelationModel(abc.ABC):
    """ MetalCorrelationModel computes stellar ages based on [Fe/H] values
    """

    def a_form(self, Fe_H, sfh, **kwargs):
        """ ages returns the ages of star particles
        """
        pass

    @abc.abstractmethod
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        pass

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out

class AbstractRanking(abc.ABC):
    """ AbstractRanking is an abstract base class for various models that 
    rank particles.
    """

    def __init__(self, ranks, core_idx, rvir, r_bins=DEFAULT_R_BINS):
        """ ranks is an array of ranks for every particle in the halo. If that
        particle has been accreted by the time of tagging, it should have an
        non-negative integer rank. If the particle hasn't been accreted yet,
        it should be set to NIL_RANK

        core_idx is the index of the core particles of the halo according to
        this ranking scheme.

        quantile_edges are the upper and lower quantile bounds for the each rank
        (i.e. the quantile, q, of a particle in rank i should be in the range
        quantile_edges[i] < q < quantile_edges[i+1]).
        """
        self.ranks = ranks
        self.n_max = np.max(ranks)
        self.core_idx = core_idx
        self.mp_star = np.zeros(len(ranks))
        
        self.idx = None
        self.x = None
        self.v = None
        self.xc = None
        self.vc = None
        self.r_bins = r_bins
        self.rvir=rvir
        self.M = None
        self.r = None
        
    def load_particles(self, x, v, idx):
        """ load_particles loads paritcle properties into the ranking. Must
        be called before mp_star() or either of the core functions. v may be
        None unless you call a later method which works with velocities (i.e.
        core_v)
        """
        self.x = np.copy(x)
        if v is not None:
            self.v = np.copy(v)
        else:
            self.v = None
        self.idx = idx

        self.xc = self.core_x()
        for dim in range(3): self.x[:,dim] -= self.xc[dim]
        
        if self.v is not None:
            self.vc = self.core_v()
            for dim in range(3): self.v[:,dim] -= self.vc[dim]

        self.r = np.sqrt(np.sum(self.x**2, axis=1))/self.rvir
        self.M = ranked_np_profile_matrix(self.ranks, self.idx,
                                          self.r, self.r_bins)

    def core_x(self):
        """ core_x returns the position of the halo core. You must have called
        load_particles() first.
        """
        if self.x is None: raise ValueError("x not set")
        core = self.x[np.searchsorted(self.idx, self.core_idx)]
        return np.median(core, axis=0)
        
    def core_v(self):
        """ core_v returns the velocity of the halo core. You must have called
        load_particles() first with a non-None argument for v.
        """
        if self.v is None: raise ValueError("v not set")
        
        core = self.v[np.searchsorted(self.idx, self.core_idx)]
        return np.median(core, axis=0)
        
    def set_mp_star(self, kwargs, profile_model, r_half, m_star):
        M = self.M
        n = M.shape[1]

        profile_kwargs = profile_model.trim_kwargs(kwargs)
        m_star_enc_target = profile_model.m_enc(
            m_star, r_half, self.r_bins[1:]*self.rvir, **profile_kwargs)
        
        dm_star_enc_target = np.zeros(len(m_star_enc_target))
        dm_star_enc_target[1:] = m_star_enc_target[1:] - m_star_enc_target[:-1]
        dm_star_enc_target[0] = m_star_enc_target[0]

        res = optimize.lsq_linear(
            M, dm_star_enc_target, bounds=(np.zeros(n), np.inf*np.ones(n))
        )
        self.mp_star_table = res.x

        self.mp_star[self.idx] = self.mp_star_table[self.ranks[self.idx]]
        is_nil = self.ranks[self.idx] == NIL_RANK
        self.mp_star[self.idx[is_nil]] = 0

        mp_star_tot = np.sum(self.mp_star)
        if mp_star_tot == 0:
            correction_frac = 1.0
        else:
            correction_frac = m_star/np.sum(self.mp_star)
        self.mp_star *= correction_frac
        self.mp_star_table *= correction_frac
        
        return self.mp_star

    def ranked_halfmass_radius(self):
        """ ranked_halfmass_radii returns an array of the current half-mass
        radii of each rank bin in pkpc.
        """
        r = np.sqrt(np.sum(self.x**2, axis=1))

        rhalf = np.zeros(self.n_max+1)
        for i in range(self.n_max+1):
            rhalf[i] = np.median(r[self.ranks[self.idx] == i])
            
        return rhalf

    def relaxation_time(self, mp, eps):
        """ relaxation_time returns the relaxation time in Gyr of each particle
        in the halo, assuming that it is on a circular orbit. mp and eps are in
        Msun and pkpc, respectively.
        """
        
        r = np.sqrt(np.sum(self.x**2, axis=1))
        
        order = np.argsort(r)
        Nr = np.zeros(len(r))
        Nr_sort = np.arange(len(Nr)) + 1
        Nr[order] = Nr_sort

        t_orb = t_orbit(Nr*mp, r)
        t_relax = t_relax_t_orbit(Nr, r, eps)*t_orb

        return t_relax
    
    def ranked_relaxation_time(self, mp, eps):
        """ ranked_relaxation_times returns an array of the relaxation times of
        each rank bin in Gyr. Ranks begin expanding when 0.18*t_relax has
        passed.
        """

        t_relax = self.relaxation_time(mp, eps)
        mean_t_relax = np.zeros(self.n_max+1)
        for i in range(len(mean_t_relax)):
            mean_t_relax[i] = np.mean(t_relax[self.ranks[self.idx] == i])
        
        return mean_t_relax
    
##################################
# Specific Model Implementations #
##################################

    
class PlummerProfile(ProfileShapeModel):
    """ PlummerProfile models a galaxy's mass distribution as a Plummer sphere.
    """
    def m_enc(self, m_star, r_half, r, **kwargs):
        """ m_enc returns the enclosed mass profile as a function of 3D radius,
        r. m_star is the asymptotic stellar mass of the galaxy, r_half is the 2D
        half-light radius of the galaxy. Returned masses will be in the same
        units as m_star.
        """
        a = r_half
        return m_star*r**3 /(r**2 + a**2)**1.5
    
    def density(self, m_star, r_half, r, **kwargs):
        """ density returns the local density as a function of 3D radius, r.
        m_star is the asymptotic stellas mass of the galaxy, r_hald is the 2D
        half-light radius of the galaxy. Returned masses will be in the same
        units of m_star.
        """
        a = r_half
        return 3*m_star/(4*np.pi*a**3) * (1 + r**2/a**2)**(-5/2)
    
    def var_names(self):
        return []

class Nadler2020RHalf(RHalfModel):
    """ Nadler20202RHalf models galaxies according to the z=0 size-mass relation
    in Nadler et al. 2020. (https://arxiv.org/abs/1912.03303)

    This is calibrated by fitting a galaxy-halo model to Milky Way satellites.
    """
    
    def __init__(self, A=27e-6, n=1.07, R0=10e-3, sigma_log_R=0.63):
        """ The constructor for Nadler2020RHalf allows you to change the
        parameters of the fit. Please consult the paper for discussion on what
        these parameters mean. This allows you to sample the posterior
        distribution for these parameters.
        """
        self.A = A
        self.n = n
        self.R0 = R0
        self.sigma_log_R = sigma_log_R

    def r_half(self, no_scatter=False, **kwargs):
        """ r_half returns the half-mass radius of a a galaxy in physical kpc.
        Required keyword arguments:
         - rvir
        """
        # inputs and outputs are in pMpc (no h).
        log_R = np.log10(self.A * (rvir/self.R0)**self.n)
        if no_scatter:
            return 10**log_R
        else:
            log_scatter = self.sigma_log_R*random.normal(
                0, 1, size=np.shape(rvir))
            return 10**(log_R + log_scatter)

    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        return ["rvir"]


class FixedRHalf(RHalfModel):
    def __init__(self, ratio=0.015):
        self.ratio = ratio

    def r_half(self, no_scatter=False, rvir=None):
        return rvir*self.ratio

    def var_names(self):
        return ["rvir"]

class Jiang2019RHalf(RHalfModel):
    """ Jiang2019Rhalf models galaxies according to the z-dependent size-mass
    relation in Jiang et al. 2019 (https://arxiv.org/abs/1804.07306; Appendix
    D). This model is fit to   galaxies in NIHAO/VELA and has been rescaled
    to  match the noramlization of the z-dependence seen in GAMA+CANDLES when
    combined with abundance matching (Somerville et al. 2018;
    https://arxiv.org/abs/1701.03526; M* > 1e8 Msun).
    """
    def __init__(self, sigma_log_R=0.2):
        """ The constructor for Jiang2019RHalf allows you to change the
        intrinsic scatter in the relation. I eyeballed sigma_log_R at 0.2 from
        their plots, so this value in particular is worth modifying.
        """
        # I eyeballed the 0.2 from their plots.
        self.sigma_log_R = sigma_log_R
    
    def r_half(self, rvir=None, cvir=None, z=None, no_scatter=False):
        """ r_half returns the half-mass radius of a a galaxy in physical kpc.
        Required keyword arguments:
         - rvir
         - cvir
         - z
        """
        if rvir is None: raise ValueError("rvir not supplied")
        if cvir is None: raise ValueError("cvir not supplied")
        if z is None: raise ValueError("z not supplied")
        
        # From Appendix D
        fz = 0.02*(1 + z)**-0.2
        R = fz * (cvir/10)**-0.7 * rvir
        log_scatter = self.sigma_log_R*random.normal(0, 1, size=np.shape(rvir))
        if not no_scatter:
            return 10**(np.log10(R) + log_scatter)
        else:
            return R

    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        return ["rvir", "cvir", "z"]

class Carlsten2021RHalf(RHalfModel):
    """ Carlsten2022RHalf models galaxies according to a z=0 size-mass relation
    calibrated off of observations in the local volume.
    (https://arxiv.org/pdf/2105.03435.pdf) (10^5.5 Msun < M* < 10^8.5 Msun).
    """
    def __init__(self, sigma_log_R=0.181, a=1.071, b=0.247):
        """ The constructor for Carlsten2021RHalf is just a power law with 
        log-normal scatter that lets yuo
        """
        self.sigma_log_R = sigma_log_R
        self.a = a
        self.b = b
    
    def r_half(self, rvir=None, cvir=None, z=None, no_scatter=False):
        """ r_half returns the half-mass radius of a a galaxy in physical kpc.
        Required keyword arguments:
         - rvir
         - cvir
         - z
        """
        R = 10**(self.a + self.b*np.log10(0.247))
        log_scatter = self.sigma_log_R*random.normal(0, 1, size=np.shape(rvir))
        if not no_scatter:
            return 10**(np.log10(R) + log_scatter)
        else:
            return R


    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        return ["rvir", "cvir", "z"]

class UniverseMachineSFH(SFHModel):
    """ SFHModel is an abstract base class for models that compute star
    formation histories.
    """
    def sfh(self, um_sfh=None):
        return um_sfh

    def var_names(self):
        return ["um_sfh"]

class DarkMatterSFH(SFHModel):
    """ SFHModel is an abstract base class for models that compute star
    formation histories.
    """
    def sfh(self, dm_mah=None):
        return dm_mah

    def var_names(self):
        return ["dm_mah"]

class Kirby2013Metallicity(FeHMeanModel):
    """ Kirby2013Metallicity models galaxy metallicity according to the
    z-agnostic fit in Kirby et al. 2013 (https://arxiv.org/pdf/1310.0814.pdf;
    Equation 4).
    """
    def __init__(self, sigma_Fe_H=0.17, Fe_H_dist_width=0.3):
        """ The constructor for Kirby2013Metallicity allows you to change the
        intrinsic scatter in the relation.
        """
        self.sigma_Fe_H = sigma_Fe_H
        #self.Fe_H_dist_width = Fe_H_dist_width
    
    def Fe_H(self, mstar=None, no_scatter=False):
        """ Fe_H returns the metallicity of a given galaxy.
        Required keyword arguments:
         - mstar
        """
        if mstar is None: raise ValueError("mstar not supplied")
        
        Fe_H_mean = -1.69 + 0.30*np.log10(mstar/1e6)
        if not no_scatter:
            Fe_H_mean = random.normal(Fe_H_mean, self.sigma_Fe_H)
        return Fe_H_mean

        #return random.normal(Fe_H_mean, self.Fe_H_dist_width, size=n_part)


    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        return ["mstar"]
    
class Kirby2013MDF(FeHMDFModel):
    def __init__(self, model_type="leaky box"):
        """ accepted model_types are:
          - "leaky box"

        Maybe I'll add pre-enriched and accretion later if people care.

        Note: the leaky box model is nice because it's a single parameter
        model, but it doesn't fit higher mass galaxies as well as more
        complicated models.
        """
        if model_type != "leaky box":
            raise ValueError("Unknown model_type, %s" % model_type)

        self.model_type = model_type

    def mdf(self, Fe_H_mean):
        if self.model_type == "leaky box":
            def leaky_mean(p_eff):
                pdf = lambda Fe_H: 10**Fe_H/p_eff*np.exp(-10**Fe_H/p_eff)
                return (integrate.quad(lambda Fe_H: Fe_H*pdf(Fe_H), -10, 5)[0] /
                        integrate.quad(lambda Fe_H: pdf(Fe_H), -10, 5)[0])
            log_p_eff = optimize.fsolve(
                lambda log_p_eff: leaky_mean(10**log_p_eff) - Fe_H_mean,
                -1.5)
            p_eff = 10**log_p_eff
            return sampling.PDF(lambda Fe_H: 10**Fe_H/p_eff*np.exp(-10**Fe_H/p_eff), -10, 5)
            
            #p_eff = 10**np.random.normal(-1.460, 0.306)
            #pdf_i = PDF(lambda Fe_H: 10**Fe_H/p_eff*np.exp(-10**Fe_H/p_eff))
            #mean_i = pdf_i.mean()
            #def shifted_pdf(Fe_H):
            #    dFe_H = Fe_H - mean_i
            #    return 
        else:
            assert(0)

    def var_names(self):
        return []

class FlatFeHProfile(FeHProfileModel):
    def Fe_H_profile(self, mdf, ranks):
        return mdf.sample(len(ranks.ranks))

    def var_names(self):
        return []

class GaussianCoupalaCorrelation(MetalCorrelationModel):
    """ GaussianCoupalaCorrelation connects ages to metallicities by assuming 
    that thjey can be represented by a Gaussian coupala.
    """

    def __init__(self, rho=0.76):
        self.rho = rho

    def a_form(self, Fe_H, sfh):
        return sampling.gaussian_coupala_sample(Fe_H, sfh, self.rho)
        
    def var_names(self):
        return []

    def trim_kwargs(self, kwargs):
        out = {}
        for key in self.var_names():
            out[key] = kwargs[key]
        return out

class UniverseMachineMStarFit(MStarModel):
    def m_star(self, mpeak=None, z=None, no_scatter=False):
        """
         Required keyword arguments:
         - mpeak
         - z
        """

        if mpeak is None: raise ValueError("mpeak not supplied")
        if z is None: raise ValueError("z not supplied")
        
        mpeak = mpeak
        
        a = 1/(1 + z)

        e0 = -1.435
        al_lna = -1.732

        ea = 1.831
        alz = 0.178

        e_lna = 1.368
        b0 = 0.482

        ez = -0.217
        ba = -0.841

        m0 = 12.035
        bz = -0.471

        ma = 4.556
        d0 = 0.411

        m_lna = 4.417
        g0 = -1.034

        mz = -0.731
        ga = -3.100

        al0 = 1.963
        gz = -1.055

        ala = -2.316
        
        log10_M1_Msun = m0 + ma*(a-1) - m_lna*np.log(a) + mz*z
        e = e0 + ea*(a - 1) - e_lna*np.log(a) + ez*z
        al = al0 + ala*(a - 1) - al_lna*np.log(a) + alz*z
        b = b0 + ba*(a - 1) + bz*z
        d = d0
        g = 10**(g0 + ga*(a - 1) + gz*z)

        x = np.log10(mpeak/10**log10_M1_Msun)
      
        log10_Ms_M1 = (e - np.log10(10**(-al*x) + 10**(-b*x)) +
                       g*np.exp(-0.5*(x/d)**2))
                       
        log10_Ms_Msun = log10_Ms_M1 + log10_M1_Msun

        if not no_scatter:
            log_scatter = 0.2*random.normal(0, 1, size=np.shape(mpeak))
            log10_Ms_Msun += log_scatter
        
        Ms = 10**log10_Ms_Msun
        
        return Ms
    
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """
        return ["mpeak", "z"]
    

class UniverseMachineMStar(MStarModel):
    def m_star(self, um_mstar=None, no_scatter=False):
        return np.maximum(1, um_mstar)

    def var_names(self):
        return ["um_mstar"]

class RadialEnergyRanking(AbstractRanking):
    def __init__(self, params, x, v, idx, n_max, rvir,
                 core_particles=DEFAULT_CORE_PARTICLES,
                 quantile_edges=DEFAULT_QUANTILE_EDGES):
        """ Takes mp (Msun; may be a scalar), x (pkpc), v (pkm/s), idx (the
        index into the full particle array), and n_max (the number of particles
        in the full particle arrays).
        
        only_tag may be set so that only a subset of particles may be tagged
        (e.g. early-accreted particles).
        """        
        (self.rmax, self.vmax,
         self.pe_vmax2, self.order) = profile_info(params, x)

        self.ke_vmax2 = 0.5*np.sum(v**2, axis=1) / self.vmax**2
        self.E_vmax2 = self.ke_vmax2 + self.pe_vmax2

        core_q = min(core_particles / len(self.E_vmax2), 1)
        core_E_vmax2 = np.quantile(self.E_vmax2, core_q)
        
        core_idx = idx[self.E_vmax2 <= core_E_vmax2]

        E_edges = np.linspace(-10, 0, 41)
        quantile_edges = [np.sum(self.E_vmax2< E_edges[i])/
                          len(self.E_vmax2) for i in range(len(E_edges))]

        ranks, self.E_edges = rank_by_quantile(
            quantile_edges, self.E_vmax2, idx, n_max)
    
        super(RadialEnergyRanking, self).__init__(ranks, core_idx, rvir)
        
#################################
# General classes and functions #
#################################

def tag_stars(sim_dir, galaxy_halo_model, star_snap=None, E_snap=None,
              target_subs=None):
    # Basic simulation information
    param = lib.simulation_parameters(sim_dir)
    h, hist = lib.read_subhalos(sim_dir)
    h_cmov, _ = lib.read_subhalos(sim_dir, comoving=True)
    scale = lib.scale_factors(sim_dir)

    # Not everyone is going to run UM on their zoom-ins, so don't force this
    # file to be read.
    if "um_mstar" in galaxy_halo_model.var_names():
        um = lib.read_um(sim_dir)
    else:
        um = [None]*len(h)

    if target_subs is None:
        # Don't read in the host: wouldn't make sense.
        target_subs = np.arange(1, len(h), dtype=int)

    if 0 in target_subs:
        raise ValueError("Cannot do star-tagging on the central halo. " + 
                         "Remove 0 from the target_subs array.")

    # Set the optional arguments to their default values
    if star_snap is None:
        # Match profiles at the snapshot when the subhalo first crosses the
        # virial radius.
        star_snap = hist["first_infall_snap"]
    if E_snap is None:
        E_snap = np.zeros(len(h), dtype=int)
        for i in target_subs:
            # Look back an eighth of an orbital time, or to the half-mass
            # scale, whichever is later.
            if star_snap[i] == 0:
                #raise ValueError(("Subhalo %d has a star-tagging snapshot " + 
                #                  "of 0. Either change star_snap for " + 
                #                  "this subhalo or remove it from " + 
                #                  "target_subs") % i)
                E_snap[i] = 0
            else:
                E_snap[i] = look_back_orbital_time(
                    param, scale, star_snap[i], 0.125, h[i,:], 0.5)

    # Set up buffers and shared information for I/O.

    # Total number of particles in each snapshot
    n_max = np.zeros(len(h), dtype=int)
    # The indices of the particles stored at the two stages. They get trimmed
    # down to only the valid and self-owned particles.
    idx_star, idx_E = [None]*len(h), [None]*len(h)
    # Position and velocity information for the different particles.
    x_star, x_E, v_E = [None]*len(h), [None]*len(h), [None]*len(h)
    # Shared information across snapshots.
    part_info = lib.ParticleInfo(sim_dir)

    for snap in range(param["n_snap"]):
        # If nobody needs this snapshot, don't read it in.
        if (snap not in E_snap[target_subs] and
            snap not in star_snap[target_subs]): continue

        # Work out which objects we actually need to read in this snapshot.
        star_ok = np.zeros(len(h), dtype=bool)
        E_ok = np.zeros(len(h), dtype=bool)
        star_ok[target_subs], E_ok[target_subs] = True, True
        star_ok = star_ok & (star_snap == snap)
        E_ok = E_ok & (E_snap == snap)
        i_star, i_E = np.where(star_ok)[0], np.where(E_ok)[0]

        # Read in data.
        host, a_z = h_cmov[0,snap], scale[snap]

        # Add information at the snapshot stellar masses are assigned
        for i in i_star:
            ok = lib.read_particles(
                part_info, sim_dir, snap, "valid", owner=i)
            x = lib.read_particles(
                part_info, sim_dir, snap, "x", owner=i)
            v = lib.read_particles(
                part_info, sim_dir, snap, "v", owner=i)

            idx_star[i] = np.where(ok)[0]
            sx = h[i,snap]["x"]
            x_star[i] = util.set_units_x(x[ok], host, a_z, param) - sx
            n_max[i] = len(x)

        # Add information at the snapshot energies are measured
        for i in i_E:
            ok = lib.read_particles(
                part_info, sim_dir, snap, "valid", owner=i)
            x = lib.read_particles(
                part_info, sim_dir, snap, "x", owner=i)
            v = lib.read_particles(
                part_info, sim_dir, snap, "v", owner=i)

            idx_E[i] = np.where(ok)[0]
            sx, sv = h[i,snap]["x"], h[i,snap]["v"]

            x_E[i] = util.set_units_x(x[ok], host, a_z, param) - sx
            v_E[i] = util.set_units_v(v[ok], host, a_z, param) - sv

    # Finally, assign stellar masses and create ranking objects for each halo.
    mp_star, ranks = [None]*len(h), [None]*len(h)

    r_half, m_star = np.ones(len(h))*-1, np.ones(len(h))*-1
    Fe_H = [None]*len(h)

    gal_hists = np.zeros(len(h), dtype=lib.GALAXY_HISTORY_DTYPE)
    stars = [None]*len(h)
    for i in target_subs:
        stars[i] = np.zeros(n_max[i], dtype=lib.STAR_DTYPE)

    for i in target_subs:
        ranks[i] = RadialEnergyRanking(
            param, x_E[i], v_E[i], idx_E[i], n_max[i],
            h["rvir"][i, star_snap[i]])
        ranks[i].load_particles(x_star[i], None, idx_star[i])

        # It's some tiny subhalo that Rockstar made a mistake on. Doesn't have
        # any bins with more than 10 particles in them.
        if np.max(ranks[i].ranks) == -1:
            continue

        kwargs = galaxy_halo_model.get_kwargs(
            param, scale, h[i], um[i], star_snap[i])

        stars[i], gal_hists[i] = galaxy_halo_model.star_properties(
            ranks[i], kwargs)

    return stars, gal_hists, ranks

class RetagStarsState(object):
    def __init__(self, sim_dir, galaxy_halo_model):
        self.param = lib.simulation_parameters(sim_dir)
        self.h, self.hist = lib.read_subhalos(sim_dir)
        self.h_cmov, _ = lib.read_subhalos(sim_dir, comoving=True)
        self.scale = lib.scale_factors(sim_dir)

        if "um_mstar" in galaxy_halo_model.var_names():
            self.um = lib.read_um(sim_dir)
        else:
            self.um = [None]*len(self.h)
    
    def get_all(self):
        return self.param, self.h, self.hist, self.h_cmov, self.scale, self.um

def retag_stars(sim_dir, galaxy_halo_model, ranks,
                state=None, star_snap=None,
                target_subs=None):
    # Basic simulation information

    if state is None:
        state = RetagStarsState(sim_dir, galaxy_halo_model)
    param, h, hist, h_cmov, scale, um = state.get_all()
    if star_snap is None:
        star_snap = hist["first_infall_snap"]
    if target_subs is None:
        # Don't read in the host: wouldn't make sense.
        target_subs = np.arange(1, len(h), dtype=int)

    if 0 in target_subs:
        raise ValueError("Cannot do star-tagging on the central halo. " + 
                         "Remove 0 from the target_subs array.")

    gal_hists = np.zeros(len(h), dtpye=lib.GALAXY_HISTORY_DTYPE)
    stars = [None]*len(h)
    for i in target_subs:
        stars[i] = np.zeros(len(ranks[i].ranks), dtype=lib.STAR_DTYPE)

    for i in target_subs:
        # It's some tiny subhalo that Rockstar made a mistake on. Doesn't have
        # any bins with more than 10 particles in them.
        if np.max(ranks[i].ranks) == -1: continue

        kwargs = galaxy_halo_model.get_kwargs(
            param, scale, h[i], um[i], star_snap[i])

        stars[i], gal_hists[i] = galaxy_halo_model.star_properties(
            ranks[i], kwargs)

    return stars, gal_hists, state            
    
def look_back_orbital_time(params, scale, snap, dt_orbit, halo, min_mass_frac):
    """ look_back_orbital_time returns the snapshot of the time which has
    experienced dt_orbit orbital times before the given snapshot, snap. To
    protect against looking back too far, you may specify a maximum amount that
    the halo can grow between the look-back snapshot. halo is an object from
    the subhalos.dat file.
    """
    cosmo = cosmology.setCosmology("", lib.colossus_parameters(params))
    
    z = 1/scale - 1
    age = cosmo.age(z)
    T_orbit = mass_so.dynamicalTime(z, "vir", "orbit")
    
    dT = (age - age[snap])
    dT_orbit = dT/T_orbit

    orbit_start = np.searchsorted(dT_orbit, -dt_orbit)

    if snap == orbit_start: return snap

    for snap_start in range(snap, orbit_start, -1):
        if halo["mvir"][snap_start-1] < min_mass_frac*halo["mvir"][snap]:
            break

    return snap_start

class StellarMassModel(object):
    def __init__(self, m_star_model, sfh_model):
        self.m_star_model = m_star_model
        self.sfh_model = sfh_model

class ProfileModel(object):
    def __init__(self, r_half_model, profile_shape_model):
        self.r_half_model = r_half_model
        self.profile_shape_model = profile_shape_model

class MetalModel(object):
    def __init__(self, Fe_H_model, Fe_H_mdf_model, Fe_H_profile_model,
                 correlation_model):
        self.Fe_H_model = Fe_H_model
        self.Fe_H_mdf_model = Fe_H_mdf_model
        self.Fe_H_profile_model = Fe_H_profile_model
        self.correlation_model = correlation_model
        
class GalaxyHaloModel(object):
    def __init__(self, stellar_mass_model, profile_model, metal_model):
        """ GalaxyHaloModel requires a model for the M*-Mhalo relation,
        m_star_model, a model for how the projected half-mass radius and Mhalo
        are related, and a model for the halo profile, profile_model. These
        should be types that inherit from AbstractMstarModel,
        AbstractRHalfModel, AbstractProfileModel, and AbstractMetallicityModel
        respectively.
        """
        self.m_star_model = stellar_mass_model.m_star_model
        self.sfh_model = stellar_mass_model.sfh_model
        self.r_half_model = profile_model.r_half_model
        self.profile_shape_model = profile_model.profile_shape_model
        self.Fe_H_model = metal_model.Fe_H_model
        self.Fe_H_mdf_model = metal_model.Fe_H_mdf_model
        self.Fe_H_profile_model = metal_model.Fe_H_profile_model
        self.correlation_model = metal_model.correlation_model
        
    def star_properties(
            self, ranks, kwargs, r_half=None, m_star=None, Fe_H=None):
        """ star_properties sets the stellar masses of a halo's dark matter
        particles given their positions relative to the halo center, and
        ranking, ranks (type: inherits from AbstractParticleRanking). This
        function accepts the same keyword arguments as its m_star_model and
        r_half_model arguments. You may also fix the half-mass radius and
        stellar mass to whatever you want with r_half (units: pkpc) and m_star
        (units: Msun).
        
        This function also returns the M_star, r_half, and [Fe/H] values that
        it assigned to the halo.
        """
        if m_star is None:
            check_var_names(kwargs, self.m_star_model)
            m_star = self.m_star_model.m_star(
                **self.m_star_model.trim_kwargs(kwargs))
            
        kwargs["mstar"] = m_star

        if r_half is None:
            check_var_names(kwargs, self.r_half_model)
            r_half = self.r_half_model.r_half(
                **self.r_half_model.trim_kwargs(kwargs))
  
        mp_star = ranks.set_mp_star(kwargs, self.profile_shape_model,
                                    r_half, m_star)
        gal_hist = np.zeros(1, dtype=lib.GALAXY_HISTORY_DTYPE)[0]

        if Fe_H is None:
            check_var_names(kwargs, self.Fe_H_model)
            Fe_H_mean = self.Fe_H_model.Fe_H(
                **self.Fe_H_model.trim_kwargs(kwargs))

        mdf = self.Fe_H_mdf_model.mdf(Fe_H_mean,
            **self.Fe_H_mdf_model.trim_kwargs(kwargs))
        sfh = self.sfh_model.sfh(**self.sfh_model.trim_kwargs(kwargs))
        Fe_H = self.Fe_H_profile_model.Fe_H_profile(mdf, ranks,
            **self.Fe_H_profile_model.trim_kwargs(kwargs))
        a_form = self.correlation_model.a_form(Fe_H, sfh,
            **self.correlation_model.trim_kwargs(kwargs))

        gal_hist["sigma_Fe_H_i"] = np.std(Fe_H)
        gal_hist["Fe_H_i"] = Fe_H_mean
        gal_hist["a50"] = sfh[0,np.searchsorted(sfh[1,:], sfh[1,-1]*0.5)]
        gal_hist["a90"] = sfh[0,np.searchsorted(sfh[1,:], sfh[1,-1]*0.9)]
        
        stars = np.zeros(len(mp_star), dtype=lib.STAR_DTYPE)
        stars["mp"] = mp_star
        stars["Fe_H"] = Fe_H
        stars["a_form"] = a_form

        gal_hist["m_star_i"] = m_star
        gal_hist["r_half_2d_i"] = r_half
        
        return stars, gal_hist
    
    def var_names(self):
        """ var_names returns the names of the variables this model requires.
        """

        models = [self.m_star_model,
                  self.sfh_model,
                  self.r_half_model,
                  self.profile_shape_model,
                  self.Fe_H_model,
                  self.Fe_H_mdf_model,
                  self.Fe_H_profile_model,
                  self.correlation_model]
        name_set = [m.var_names() for m in models]

        out_dict = { }
        for n in name_set:
            for i in range(len(n)):
                out_dict[n[i]] = None
            
        if "rvir" not in out_dict: out_dict["rvir"] = None
        return sorted(list(out_dict.keys()))

    def get_kwargs(self, params, scale, halo, um, snap):
        """ get_kwargs creates the kwargs that that are needed for a call to
        set_mp_star.
        """
        h100 = params["h100"]
        
        z = 1/scale[snap] - 1
        
        var_names = self.var_names()

        kwargs = { }
        
        for i in range(len(var_names)):
            if var_names[i] == "z":
                kwargs["z"] = z
                continue

            if var_names[i] == "mpeak":
                x = np.max(halo["mvir"][:snap+1])
            elif var_names[i] == "mstar":
                continue
            elif var_names[i] == "um_mstar":
                x = um[snap]["m_star"]
            elif var_names[i] == "dm_mah":
                x = np.zeros((2, len(scale)))
                x[0,:] = scale
                x[1,:] = running_mpeak(halo["mvir"])
            elif var_names[i] == "um_sfh":
                x = np.zeros((2, len(scale)))
                x[0,:] = scale
                x[1,:] = np.cumsum(um["sfr"]) / np.sum(um["sfr"]) * np.max(um["m_star"])
            else:
                x = halo[snap][var_names[i]]
            
            kwargs[var_names[i]] = x

        return kwargs
    
def running_mpeak(m):
    out = np.zeros(m.shape)
    out[0] = m[0]
    for i in range(1, len(m)):
        out[i] = max(out[i-1], m[i])
    return out

def check_var_names(kwargs, model):
    """ check_var_names checks whether kwargs contains all the arguments that
    model requires.
    """
    model_names = model.var_names()
    for i in range(len(model_names)):
        if model_names[i] not in kwargs:
            raise ValueError(("The variable '%s' is required, but was not " +
                              "given as an argument to set_mp_star.") %
                             model_names[i])
    
        
def clean_particles(params, x, v, s, scale, ok=None):
    """ clean_particles performs various normalizing and centering operations
    on particle positions and velocities. x and v are the positions and
    velocities of the particles in Gadget code units (cMpc/h, ckm/s). h is a
    halo from a merger.dat file. h100 is H0 / (100 km/s/Mpc) and scale is the
    scale factor. vp can be None and will be ignored if so.

    Returns x_clean, v_clean, idx. Particles that have not yet been accreted
    or which are False in the ok array will be removed. x_clean are centered
    on the halo and in units of pkpc. v_clean are also centered on the halo and
    are in units of pkm/s. idx is the indices of x_clean and v_clean in the
    original x and v arrays.
    """

    h100 = params["h100"]
    
    if ok is None:
        ok = x[:,0] > 0
    else:
        ok = ok & (x[:,0] > 0)

    x = np.copy(x)
    if v is not None: v = np.copy(v)
        
    for dim in range(3):
        if v is not None:
            v[:,dim] *= np.sqrt(scale)
            v[:,dim] -= s["v"][dim]
        x[:,dim] -= s["x"][dim]
        x[:,dim] *= scale
        
    idx = np.arange(len(x))
    x, idx = x[ok], idx[ok]
    if v is not None:
        v = v[ok]

    return 1e3*x/h100, v, idx

def v_circ(m, r):
    """ v_circ returns the circular velocity (in km/s) for a given mass (Msun)
    and radius (pkpc)
    """
    return 655.8 * (m/1e14)**0.5 * (r/1e3)**-0.5

def rmax_vmax(params, x, ok=None, bins=300):
    eps = params["eps"]/params["h100"]
    mp, h100 = params["mp"], params["h100"]
    mp /= h100
    
    if ok is None:
        r = np.sqrt(np.sum(x**2, axis=1))
    else:
        r = np.sqrt(np.sum(x[ok]**2, axis=1))

    log_r_min = np.log10(params["eps"]/params["h100"])
    r[r < eps/10] = eps/10
    log_r_max = np.log10(np.max(r))
    d_log_r = (log_r_max - log_r_min)/bins

    log_r = np.log10(r)
    ri = np.asarray(np.floor(((log_r - log_r_min)/d_log_r)), dtype=int)
    ri[ri < -1] = -1
    ri += 1
    
    n = np.bincount(ri)
    mass = np.cumsum(n*mp)
    r = 10**(np.arange(len(n))*d_log_r + log_r_min)
    Vr = v_circ(mass, r)
    
    i_max = np.argmax(Vr) 
    return r[i_max], Vr[i_max]


def profile_info(params, x, ok=None, order=None):
    """ profile_info returns basic information about the spherically averaged
    profile of a halo. x is the the position of particles (pkpc) relative to
    the halo center, mp is the particle mass[es] (Msun), and ok flags 
    which particles should be used ignored (e.g. if you're iteratively removing
    particles after unbinding). Set to None if not needed.

    returns r_max, v_max, PE/Vmax, and the order of particles according to
    their radii.
    """

    mp, h100 = params["mp"], params["h100"]
    mp /= h100
    
    r = np.sqrt(np.sum(x**2, axis=1))

    if ok is not None:
        r[np.isnan(r)] = np.max(r[~np.isnan(r)])
    
    if order is None: order = np.argsort(r)
    if ok is not None and np.sum(ok) == 0:
        return 0.0, 0.0, np.zeros(len(x)), order
    
    r_sort = np.sqrt(r[order]**2 + (params["eps"])**2)
    dm = np.ones(len(r_sort))*mp
    if ok is not None: dm[~ok] = 0
    m_enc = np.cumsum(dm[order])

    np.sum(np.isnan)
    v_rot = v_circ(m_enc, r_sort)
    i_max = np.argmax(v_rot)
    rmax, vmax = r_sort[i_max], v_rot[i_max]
    
    dW = np.zeros(len(m_enc))

    dr = r_sort[1:] - r_sort[:-1]
    dr[dr == 0] = np.finfo(dr.dtype).eps
    r_scaled = (r_sort[1:]*r_sort[:-1]) / dr
    dW[:-1] = v_circ(m_enc[:-1], r_scaled)**2

    # I don't think this is true when there's a tidal radius. Come back to this
    # later.
    vesc_lim = v_circ(m_enc[-1], r_sort[-1]) * np.sqrt(2)

    W = (np.cumsum(dW[::-1])[::-1] + 2*vesc_lim**2)/vmax**2
    out = np.zeros(len(W))
    out[order] = W
    
    return rmax, vmax, -out, order

def rank_by_quantile(quantiles, x, idx, n_max):
    """ rank_by_quantile breaks the paritcles represented by x (any value) and
    idx (index within the full set of particles, of length n_max). quantiles
    gives the upper and lower quantile edges for each rank.
    """
    q = np.quantile(x, quantiles)
    valid_edges = np.zeros(len(q), dtype=bool)
    ranks = np.ones(n_max, dtype=int)*NIL_RANK

    n_valid_edges, curr_rank = 0, 0
    for i in range(len(q) - 1):
        ok = (ranks[idx] == -1) & (x < q[i+1])
        if np.sum(ok) < MIN_PARTICLES_PER_QUANTILE: continue
        ranks[idx[ok]] = curr_rank
        curr_rank += 1
        
        valid_edges[i+1] = True
        if n_valid_edges == 0: valid_edges[i] = True
        n_valid_edges += 1
        
    return ranks, q[valid_edges]

def accreted_early(curr_snap, accrete_snap, f_accrete):
    accreted_so_far = accrete_snap <= curr_snap
    mid_snap = np.quantile(accrete_snap[accreted_so_far], f_accrete)
    return accrete_snap <= mid_snap

def t_relax_t_orbit(Nr, r, eps):
    return Nr/4 * (np.log(r**2/eps**2 + 1) +
                   eps**2-2*r**2/(3*(eps**2+r**2)) -
                   np.log(3/2))**-1

def t_orbit(Mr, r):
    return 7.5 * (r/40)**1.5 * (1e10/Mr)**0.5

def ranked_np_profile_matrix(ranks, idx, r, bins):
    n_rank = 1 + np.max(ranks)
    n_bins = len(bins)

    M = np.zeros((n_bins-1, n_rank))

    for i in range(n_rank):
        ri = r[ranks[idx] == i]
        N, _ = np.histogram(ri, bins=bins)
        #M[:,i] =  np.cumsum(N)
        M[:,i] = N
        
    return M


DWARF_GALAXY_HALO_MODEL = GalaxyHaloModel(
    StellarMassModel(
        UniverseMachineMStar(),
        UniverseMachineSFH()
    ),
    ProfileModel(
        Jiang2019RHalf(),
        PlummerProfile()
    ),
    MetalModel(
        Kirby2013Metallicity(),
        Kirby2013MDF(model_type="leaky box"),
        FlatFeHProfile(),
        GaussianCoupalaCorrelation()
    )
)

DWARF_GALAXY_HALO_MODEL_NO_UM = GalaxyHaloModel(
    StellarMassModel(
        UniverseMachineMStarFit(),
        DarkMatterSFH()
    ),
    ProfileModel(
        Jiang2019RHalf(),
        PlummerProfile()
    ),
    MetalModel(
        Kirby2013Metallicity(),
        Kirby2013MDF(model_type="leaky box"),
        FlatFeHProfile(),
        GaussianCoupalaCorrelation()
    )
)



from . import util
from . import lib
from os import path
import os
import shutil
import requests
import time

DOWNLOAD_TARGETS = ["halos", "trees", "particles", "full_snapshots"]
BASE_URL_FORMAT = "https://%s:%s@s3df.slac.stanford.edu/data/kipac/symphony/"
RETRY_WAIT_TIME = 5

def pack_files(suite, halo_name, base_out_dir, target="halos",
               logging=True):
    if logging:
        print("Packing halo %s in suite %s to %s" % (
            str(halo_name), suite, base_out_dir))
    
    if suite is None:
        for suite in ["SymphonyLMC", "SymphonyMilkyWay", "SymphonyGroup",
                      "SymphonyLCluster", "SymphonyCluster"]:
            pack_files(suite, halo_name, base_out_dir, target=target)

    if halo_name is None:
        for i in range(util.n_hosts(suite)):
            pack_files(suite, i, base_out_dir, target=target)
        return

    # MUST BE RUN FROM base_dir!!!
    if isinstance(halo_name, int):
        halo_name = util.DEFAULT_HALO_NAMES[suite][halo_name]

    if target == "trees":
        rel_dir = path.join(suite, halo_name)
        rel_out = path.join(base_out_dir, suite, target)
        os.system("mkdir -p %s" % rel_out)
        os.system("tar --exclude='%s/halos/*core*' --exclude='%s/particles' -cf %s/%s.tar.gz %s" % (rel_dir, rel_dir, rel_out, halo_name, rel_dir))
    elif target == "halos":
        rel_dir = path.join(suite, halo_name)
        rel_out = path.join(base_out_dir, suite, target)
        os.system("mkdir -p %s" % rel_out)
        os.system("tar --exclude='%s/halos/*core*' --exclude='%s/halos/tree_[0-9].dat' --exclude='%s/particles' -cf %s/%s.tar.gz %s" % (rel_dir, rel_dir, rel_dir, rel_out, halo_name, rel_dir))


        
def download_packed_files(user, password, suite, halo_name, base_out_dir,
                          target="halos", logging=True, retries=5):
    orig_halo_name = halo_name
    if logging:
        print("Downloading halo %s in suite %s to %s" % (
            str(halo_name), suite, base_out_dir))
    
    if suite is None:
        for suite in ["SymphonyLMC", "SymphonyMilkyWay", "SymphonyGroup",
                      "SymphonyLCluster", "SymphonyCluster"]:
            download_packed_files(user, password, suite, halo_name, base_out_dir,
                                  target=target, logging=logging, retries=retries)
        return
        
    if halo_name is None:
        for i in range(util.n_hosts(suite)):
            download_packed_files(user, password, suite, i, base_out_dir,
                                  target=target, logging=logging, retries=retries)
        return

    if target not in DOWNLOAD_TARGETS:
        raise ValueError("Unrecoginzed download target, %s" % target)
    
    if suite not in lib.parameter_table:
        raise ValueError("Unrecognized simulation suite, %s" % suite)
    
    if isinstance(halo_name, int):
        halo_name = util.DEFAULT_HALO_NAMES[suite][halo_name]

    if halo_name not in util.DEFAULT_HALO_NAMES[suite]:
        raise ValueError("Unrecognized halo, %s, for the suite %s" %
                         (str(halo_name), suite))
                
    os.makedirs(base_out_dir, exist_ok=True)

    base_url = BASE_URL_FORMAT % (user, password)
    base_name = "%s_%s_%s.tar" % (suite, halo_name, target)
    url = base_url + base_name
    out_file = path.join(base_out_dir, base_name)    
    # The next three lines are adapted from
    # https://stackoverflow.com/a/39217788
    # Original author is John Zwink, edited by the users vog, Martjin Peters, and
    # Daniel F.
    with requests.get(url, stream=True) as r:
        with open(out_file, 'wb') as f:
            shutil.copyfileobj(r.raw, f)
            
    if not path.exists(out_file):
        raise FileNotFoundError("Unable to download halo %s from the suite %s" %
                                (halo_name, suite))
 
    with open(out_file, "rb") as f:
        text = f.read(6)
        if text == b"<html>":
            f.seek(0, 0)
            text = f.read()
            os.remove(out_file)

            if text == b"<html>\r\n<head><title>401 Authorization Required</title></head>\r\n<body>\r\n<center><h1>401 Authorization Required</h1></center>\r\n<hr><center>nginx</center>\r\n</body>\r\n</html>\r\n":
                raise ValueError("Unrecognized user and password combination.")
            elif text == b"<html>\r\n<head><title>502 Bad Gateway</title></head>\r\n<body>\r\n<center><h1>502 Bad Gateway</h1></center>\r\n<hr><center>nginx</center>\r\n</body>\r\n</html>\r\n":
                if retries == 0:
                    raise ValueError("Recieved network error after exhausing maximum number of retries for halo %s. You could try increasing the retries variable, but the server is probably overloaded right now. Please try again later" % orig_halo_name)
                else:
                    if logging:
                        print("Internal server error ('502 Bad Gateway') encountered while downloading halo %s: trying to download again after a short wait (tries left: %d)" % (orig_halo_name, retries))
                    time.sleep(RETRY_WAIT_TIME)
                    download_packed_files(
                        user, password, suite, halo_name, base_out_dir,
                        target=target, logging=logging, retries=retries-1)
            else:
                raise ValueError("Server sent back the following error on halo %s:\n%s" %
                                 (orig_halo_name, text))
        
            
def unpack_files(suite, halo_name, base_out_dir,
                 target="halos", logging=True):
    if logging:
        print("Unpacking halo %s in suite %s to %s" % (
            str(halo_name), suite, base_out_dir))
    
    if suite is None:
        for suite in ["SymphonyLMC", "SymphonyMilkyWay", "SymphonyGroup",
                      "SymphonyLCluster", "SymphonyCluster"]:
            unpack_files(suite, halo_name, base_out_dir, target=target)
        return
            
    if halo_name is None:
        for i in range(util.n_hosts(suite)):
            unpack_files(suite, i, base_out_dir, target=target)
        return

    if isinstance(halo_name, int):
        halo_name = util.DEFAULT_HALO_NAMES[suite][halo_name]

    tar_loc = path.join(base_out_dir, "%s_%s_%s.tar" % (suite, halo_name, target))
    if not path.exists(tar_loc):
        raise FileNotFoundError("The file associated with the requested target data for host %s in suite %s does not exist within the base directory %s. This is probably because it was already downloaded or already unpacked." % (halo_name, suite, base_out_dir))
    shutil.unpack_archive(tar_loc, base_out_dir)
    os.remove(tar_loc)
    
    
def download_files(user, password, suite, halo_name, base_out_dir,
                   target="halos", logging=True, retries=5):
    download_packed_files(user, password, suite, halo_name,
                          base_out_dir, target, logging, retries=retries)
    unpack_files(suite, halo_name, base_out_dir, target, logging)



import matplotlib.pyplot as plt
import numpy as np
import os.path as path

def plot_circle(ax, x0, y0, r, **kwargs):
    """ plot_circle plots a circle with radius r centered on the position
    (x0, y0) onto the matplotlib axis, ax. This function accepts all the named
    arguments that matplotlib's plot function accepts.
    """
    th = np.linspace(0, 2*np.pi, 100)
    y = np.sin(th)*r + y0
    x = np.cos(th)*r + x0
    ax.plot(x, y, **kwargs)

def set_units_x(x, h, scale, param):
    """ set_units_x converts an array of positions with Gadget-2 code units
    to physical kpc, centered on the subhalo h (a single element of a
    SUBHALO_DTYPE array).

    h must be in comoving units.
    """
    if h["x"][0] == 0: 
        raise ValueError("The centering halo, h, must be in comoving units.")

    dx = np.zeros(x.shape)
    for dim in range(3): dx[:,dim] = x[:,dim] - h["x"][dim]
    dx *= scale*1e3/param["h100"]
    return dx

def set_units_v(v, h, scale, param):
    """ set_units_v converts an array of velocities with Gadget-2 code units
    to physical km/s, centered on the subhalo h (a single element of a
    SUBHALO_DTYPE array).

    h must be in comoving units.
    """
    if h["x"][0] == 0: 
        raise ValueError("The centering halo, h, must be in comoving units.")

    v *= np.sqrt(scale)
    dv = np.zeros(v.shape)
    for dim in range(3): dv[:,dim] = v[:,dim] - h["v"][dim]
    return dv

def set_units_parameters(scale, param):
    """ set_units_parameters returns particle mass, mp, and Plummer-equivalent
    force-softening scales in units of Msun and physical kpc, respectively.
    scale is the current scale factor and param is the parameter dictionary
    returned by simulation_parameters.
    """
    return param["mp"]/param["h100"], param["eps"]*scale/param["h100"]

def set_units_histories(hist, scale, param):
    """ set_units_histories takes an array of type HISTORY_DTYPE and returns a
    copy with units that have been converted from Rockstar's detail units to
    symlib's default units: Msun, physical kpc, and physical km/s.
    """
    hist = np.copy(hist)
    hist["mpeak"] /= param["h100"]
    hist["mpeak_pre"] /= param["h100"]
    return hist

def set_units_halos(h, scale, param):
    """ set_units_halos takes a array of type SUBHALO_DTYPE and returns a copy
    with units that have been converted from Rockstar's default units to
    symlib's default units, Msun, physical kpc centered on the host halo,
    and physical km/s centered on the host halo.
    """
    h = np.copy(h)    
    for dim in range(3):
        x0 = np.copy(h["x"][0,:,dim])
        v0 = np.copy(h["v"][0,:,dim])

        if x0[-1] == 0:
            raise ValueError("Calling set_units_halos() on a set of halos " +
                             "already in physical units.")

        for hi in range(len(h)):
            h["x"][hi,:,dim] -= x0
            h["v"][hi,:,dim] -= v0
                
            h["x"][hi,:,dim] *= 1e3*scale/param["h100"]
            
            if dim == 0:
                h["rvir"][hi,:] *= 1e3*scale/param["h100"]
                h["mvir"][hi,:] *= 1/param["h100"]
                h["rvmax"][hi,:] *= scale/param["h100"]
    
    for hi in range(len(h)):
        invalid = np.where(~h[hi,:]["ok"])[0]
        h["mvir"][hi,invalid] = -1
        h["rvir"][hi,invalid] = -1
        h["rvmax"][hi,invalid] = -1
        h["x"][hi,invalid] = -1
        h["v"][hi,invalid] = -1
        h["vmax"][hi,invalid] = -1
        h["cvir"][hi,invalid] = -1
        
    return h

DEFAULT_HALO_NAMES = {
    "SymphonyLMC": sorted([
        "Halo032", "Halo097", "Halo218", "Halo374", "Halo463", "Halo567", "Halo721", "Halo853",
        "Halo059", "Halo104", "Halo296", "Halo380", "Halo4662", "Halo575", "Halo767", "Halo914",
        "Halo0662", "Halo110", "Halo301", "Halo391", "Halo511", "Halo602", "Halo802", "Halo932",
        "Halo083", "Halo202", "Halo303", "Halo405", "Halo524", "Halo697", "Halo824", "Halo933",
        "Halo088", "Halo208", "Halo340", "Halo440", "Halo539", "Halo711", "Halo850"
    ]),
    "SymphonyMilkyWay": sorted([
        "Halo023", "Halo268", "Halo364", "Halo440", "Halo558", "Halo641", "Halo797", "Halo878", "Halo939",
        "Halo088", "Halo270", "Halo374", "Halo460", "Halo567", "Halo675", "Halo800", "Halo881", "Halo967",
        "Halo119", "Halo288", "Halo414", "Halo469", "Halo570", "Halo718", "Halo825", "Halo925", "Halo9749",
        "Halo188", "Halo327", "Halo415", "Halo490", "Halo606", "Halo738", "Halo829", "Halo926", "Halo9829",
        "Halo247", "Halo349", "Halo416", "Halo530", "Halo628", "Halo749", "Halo852", "Halo937", "Halo990"
    ]),
    "EDEN_MilkyWay_8K": sorted([
        "Halo023", "Halo268", "Halo364", "Halo440", "Halo558", "Halo641", "Halo797", "Halo878", "Halo939",
        "Halo088", "Halo270", "Halo374", "Halo460", "Halo567", "Halo675", "Halo800", "Halo881", "Halo967",
        "Halo119", "Halo288", "Halo414", "Halo469", "Halo570", "Halo718", "Halo825", "Halo925", "Halo9749",
        "Halo188", "Halo327", "Halo415", "Halo490", "Halo606", "Halo738", "Halo829", "Halo926", "Halo9829",
        "Halo247", "Halo349", "Halo416", "Halo530", "Halo628", "Halo749", "Halo852", "Halo937", "Halo990",
    ]) + ["Halo119_m", "Halo364_m", "Halo460_m", "Halo416_m", "Halo530_m", "Halo675_m", "Halo800_m", "Halo829_m", "Halo852_m"],
    "SymphonyMilkyWayDisk": sorted([
        "Halo023", "Halo268", "Halo364", "Halo440", "Halo558", "Halo641", "Halo797", "Halo878", "Halo939",
        "Halo088", "Halo270", "Halo374", "Halo460", "Halo567", "Halo675", "Halo800", "Halo881", "Halo967",
        "Halo119", "Halo288", "Halo414", "Halo469", "Halo570", "Halo718", "Halo825", "Halo925", "Halo9749",
        "Halo188", "Halo327", "Halo415", "Halo490", "Halo606", "Halo738", "Halo829", "Halo926", "Halo9829",
        "Halo247", "Halo349", "Halo416", "Halo530", "Halo628", "Halo749", "Halo852", "Halo937", "Halo990"
    ]),
    "SymphonyMilkyWayDiskDMO": sorted([
        "Halo023", "Halo268", "Halo364", "Halo440", "Halo558", "Halo641", "Halo797", "Halo878", "Halo939",
        "Halo088", "Halo270", "Halo374", "Halo460", "Halo567", "Halo675", "Halo800", "Halo881", "Halo967",
        "Halo119", "Halo288", "Halo414", "Halo469", "Halo570", "Halo718", "Halo825", "Halo925", "Halo9749",
        "Halo188", "Halo327", "Halo415", "Halo490", "Halo606", "Halo738", "Halo829", "Halo926", "Halo9829",
        "Halo247", "Halo349", "Halo416", "Halo530", "Halo628", "Halo749", "Halo852", "Halo937", "Halo990"
    ]),

    "SymphonyMilkyWayLR": sorted([
        "Halo023", "Halo247", "Halo268", "Halo530", "Halo829"
    ]),

    "SymphonyMilkyWayHR": sorted([
        "Halo023", "Halo247", "Halo268", "Halo530", "Halo829"
    ]),    

    "SymphonyGroup": sorted([
        "Halo015", "Halo175", "Halo313", "Halo352", "Halo470", "Halo579", "Halo755", "Halo806", "Halo985",
        "Halo024", "Halo183", "Halo331", "Halo383", "Halo496", "Halo581", "Halo759", "Halo834",
        "Halo029", "Halo186", "Halo336", "Halo384", "Halo501", "Halo593", "Halo774", "Halo861",
        "Halo046", "Halo257", "Halo342", "Halo399", "Halo504", "Halo606", "Halo781", "Halo909",
        "Halo055", "Halo286", "Halo346", "Halo428", "Halo551", "Halo656", "Halo785", "Halo927",
        "Halo090", "Halo302", "Halo347", "Halo444", "Halo570", "Halo752", "Halo790", "Halo962",
    ]),
    "SymphonyLCluster": sorted([
        "Halo_000", "Halo_005", "Halo_012", "Halo_017", "Halo_023", "Halo_028", "Halo_046",
        "Halo_001", "Halo_006", "Halo_013", "Halo_019", "Halo_024", "Halo_029", "Halo_047",
        "Halo_002", "Halo_008", "Halo_014", "Halo_020", "Halo_025", "Halo_042", "Halo_050",
        "Halo_003", "Halo_009", "Halo_015", "Halo_021", "Halo_026", "Halo_043",
        "Halo_004", "Halo_010", "Halo_016", "Halo_022", "Halo_027", "Halo_044",
    ]),
    "SymphonyCluster": sorted(['Halo156', 'Halo175', 'Halo200', 'Halo211', 'Halo213', 'Halo222', 'Halo225', 'Halo266', 'Halo274', 'Halo277', 'Halo282', 'Halo293', 'Halo304', 'Halo306', 'Halo308', 'Halo317', 'Halo321', 'Halo324', 'Halo326', 'Halo335', 'Halo337', 'Halo339', 'Halo345', 'Halo346', 'Halo348', 'Halo349', 'Halo352', 'Halo354', 'Halo358', 'Halo360', 'Halo361', 'Halo366', 'Halo367', 'Halo377', 'Halo378', 'Halo385', 'Halo386', 'Halo387', 'Halo390', 'Halo391', 'Halo394', 'Halo400', 'Halo407', 'Halo409', 'Halo415', 'Halo428', 'Halo437', 'Halo441', 'Halo445', 'Halo447', 'Halo448', 'Halo454', 'Halo455', 'Halo456', 'Halo461', 'Halo472', 'Halo474', 'Halo475', 'Halo476', 'Halo478', 'Halo479', 'Halo480', 'Halo483', 'Halo489', 'Halo494', 'Halo502', 'Halo517', 'Halo522', 'Halo544', 'Halo545', 'Halo546', 'Halo561', 'Halo572', 'Halo574', 'Halo600', 'Halo629', 'Halo639', 'Halo645', 'Halo734', 'Halo372', 'Halo425']),
    "SymphonyClusterCorrupted": sorted([
        "Halo156", "Halo293", "Halo339", "Halo367", "Halo409", "Halo448", "Halo475", "Halo522", "Halo629",
        "Halo175", "Halo304", "Halo345", "Halo377", "Halo415", "Halo452", "Halo476", "Halo529", "Halo631",
        "Halo200", "Halo305", "Halo346", "Halo378", "Halo416", "Halo454", "Halo478", "Halo544", "Halo639",
        "Halo211", "Halo306", "Halo348", "Halo385", "Halo419", "Halo455", "Halo479", "Halo545", "Halo645",
        "Halo213", "Halo308", "Halo349", "Halo386", "Halo428", "Halo456", "Halo480", "Halo546", "Halo653",
        "Halo222", "Halo317", "Halo352", "Halo387", "Halo429", "Halo461", "Halo483", "Halo561", "Halo734",
        "Halo225", "Halo321", "Halo354", "Halo390", "Halo436", "Halo462", "Halo489", "Halo572",
        "Halo266", "Halo324", "Halo358", "Halo391", "Halo437", "Halo465", "Halo494", "Halo574",
        "Halo274", "Halo326", "Halo360", "Halo394", "Halo441", "Halo471", "Halo502", "Halo595",
        "Halo277", "Halo335", "Halo361", "Halo400", "Halo445", "Halo472", "Halo517", "Halo600",
        "Halo282", "Halo337", "Halo366", "Halo407", "Halo447", "Halo474", "Halo518", "Halo604", "Halo372", "Halo425"
    ]),
    "MWest": sorted([
        "Halo004", "Halo170", "Halo282", "Halo407", "Halo523", "Halo666", "Halo756", "Halo908", "Halo983",
        "Halo113", "Halo222", "Halo327", "Halo453", "Halo719", "Halo788",
        "Halo169", "Halo229", "Halo349", "Halo476", "Halo659", "Halo747", "Halo858", "Halo975"
    ]),
}

SUITE_NAMES = ["SymphonyLMC", "SymphonyMilkyWay", "SymphonyMilkyWayLR", "SymphonyMilkyWayHR", "SymphonyGroup", "SymphonyLCluster", "SymphonyCluster", "SymphonyClusterCorrupted", "MWest", "EDEN_MilkyWay_8K"]

def n_hosts(suite_name):
    """ n_halos returns the number of host halos in a given suite,
    """
    if suite_name not in DEFAULT_HALO_NAMES.keys():
        raise ValueError("Unrecognized suite name, '%s'" % suite_name)
    
    return len(DEFAULT_HALO_NAMES[suite_name])

def get_host_directory(base_dir, suite_name, halo_name):
    """ get_host_directory returns the name of the directory containing a given
    halo. base_dir is the central directory, suite_name is the name of the
    suite, and halo_name is the name of the halo. halo_name can either be
    a string with the exact name of the halo or it can be an integer, indexing
    into the list of halo names in sorted order. When looping over all the
    halos in a suite, you'll want to use this second option.
    """
    
    if len(suite_name) > 2 and suite_name[-2:] == "LR":
        suite_dir = suite_name[:-2]
    else:
        suite_dir = suite_name

    if isinstance(halo_name, int) or isinstance(halo_name, np.int64) or isinstance(halo_name, np.int32):
        halo_name = DEFAULT_HALO_NAMES[suite_name][halo_name]

    return path.join(base_dir, suite_dir, halo_name)
